"""polymorthic run_metadata [8ed03137cacc].

Revision ID: 8ed03137cacc
Revises: a91762e6be36
Create Date: 2023-11-17 16:19:34.281692

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlmodel import Session

# revision identifiers, used by Alembic.
revision = "8ed03137cacc"
down_revision = "a91762e6be36"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("resource_type", sa.VARCHAR(length=255), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "resource_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.drop_constraint(
            "fk_run_metadata_step_run_id_step_run", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_run_metadata_artifact_version_id_artifact_version",
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            "fk_run_metadata_pipeline_run_id_pipeline_run", type_="foreignkey"
        )

    with Session(bind=op.get_bind()) as session:
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET resource_id = pipeline_run_id,
                resource_type = 'pipeline_run'
            WHERE pipeline_run_id IS NOT NULL
            """
        )
        session.execute(stmt)
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET resource_id = step_run_id,
                resource_type = 'step_run'
            WHERE step_run_id IS NOT NULL
            """
        )
        session.execute(stmt)
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET resource_id = artifact_version_id,
                resource_type = 'artifact_version'
            WHERE artifact_version_id IS NOT NULL
            """
        )
        session.execute(stmt)
        session.commit()

    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.alter_column(
            "resource_type",
            existing_type=sa.VARCHAR(length=255),
            existing_nullable=True,
            nullable=False,
        )
        batch_op.alter_column(
            "resource_id",
            existing_type=sqlmodel.sql.sqltypes.GUID(),
            existing_nullable=True,
            nullable=False,
        )
        batch_op.drop_column("step_run_id")
        batch_op.drop_column("artifact_version_id")
        batch_op.drop_column("pipeline_run_id")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("pipeline_run_id", sa.CHAR(length=32), nullable=True)
        )
        batch_op.add_column(
            sa.Column("artifact_version_id", sa.CHAR(length=32), nullable=True)
        )
        batch_op.add_column(
            sa.Column("step_run_id", sa.CHAR(length=32), nullable=True)
        )
        batch_op.create_foreign_key(
            "fk_run_metadata_pipeline_run_id_pipeline_run",
            "pipeline_run",
            ["pipeline_run_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            "fk_run_metadata_artifact_version_id_artifact_version",
            "artifact_version",
            ["artifact_version_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            "fk_run_metadata_step_run_id_step_run",
            "step_run",
            ["step_run_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with Session(bind=op.get_bind()) as session:
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET pipeline_run_id = resource_id
            WHERE resource_type = 'pipeline_run'
            """
        )
        session.execute(stmt)
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET step_run_id = resource_id
            WHERE resource_type = 'step_run'
            """
        )
        session.execute(stmt)
        stmt = sa.text(
            """
            UPDATE run_metadata
            SET artifact_version_id = resource_id
            WHERE resource_type = 'artifact_version'
            """
        )
        session.execute(stmt)
        session.commit()

    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.drop_column("resource_id")
        batch_op.drop_column("resource_type")

    # ### end Alembic commands ###



"""Add identity table [7e4a481d17f7].

Revision ID: 7e4a481d17f7
Revises: 0.30.0
Create Date: 2022-12-08 16:27:48.909015

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "7e4a481d17f7"
down_revision = "0.30.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    from zenml.config.global_config import GlobalConfiguration

    op.create_table(
        "identity",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # get metadata from current connection
    meta = sa.MetaData()

    # pass in tuple with tables we want to reflect, otherwise whole database
    #  will get reflected
    meta.reflect(only=("identity",), bind=op.get_bind())
    # Prefill the identity table with a single row that contains the deployment
    # id extracted from the global configuration
    op.bulk_insert(
        sa.Table(
            "identity",
            meta,
        ),
        [
            {
                "id": str(GlobalConfiguration().user_id).replace("-", ""),
            },
        ],
    )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    op.drop_table("identity")



"""Release [0.42.0].

Revision ID: 0.42.0
Revises: 0e4735b23577
Create Date: 2023-07-20 09:46:06.914807

"""

# revision identifiers, used by Alembic.
revision = "0.42.0"
down_revision = "0e4735b23577"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of ZenML's UUID materializer."""

import hashlib
import os
import uuid
from typing import Any, ClassVar, Dict, Optional, Tuple, Type

from zenml.artifact_stores.base_artifact_store import BaseArtifactStore
from zenml.enums import ArtifactType
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.metadata.metadata_types import MetadataType

DEFAULT_FILENAME = "uuid.txt"


class UUIDMaterializer(BaseMaterializer):
    """Materializer to handle UUID objects."""

    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (uuid.UUID,)
    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA

    def __init__(
        self, uri: str, artifact_store: Optional[BaseArtifactStore] = None
    ):
        """Define `self.data_path`.

        Args:
            uri: The URI where the artifact data is stored.
            artifact_store: The artifact store where the artifact data is stored.
        """
        super().__init__(uri, artifact_store)
        self.data_path = os.path.join(self.uri, DEFAULT_FILENAME)

    def load(self, _: Type[uuid.UUID]) -> uuid.UUID:
        """Read UUID from artifact store.

        Args:
            _: The type of the data to be loaded.

        Returns:
            The loaded UUID.
        """
        with self.artifact_store.open(self.data_path, "r") as f:
            uuid_str = f.read().strip()
        return uuid.UUID(uuid_str)

    def save(self, data: uuid.UUID) -> None:
        """Write UUID to artifact store.

        Args:
            data: The UUID to be saved.
        """
        with self.artifact_store.open(self.data_path, "w") as f:
            f.write(str(data))

    def extract_metadata(self, data: uuid.UUID) -> Dict[str, MetadataType]:
        """Extract metadata from the UUID.

        Args:
            data: The UUID to extract metadata from.

        Returns:
            A dictionary of metadata extracted from the UUID.
        """
        return {
            "string_representation": str(data),
        }

    def compute_content_hash(self, data: uuid.UUID) -> Optional[str]:
        """Compute the content hash of the given data.

        Args:
            data: The data to compute the content hash of.

        Returns:
            The content hash of the given data.
        """
        hash_ = hashlib.md5(usedforsecurity=False)
        hash_.update(self.__class__.__name__.encode())
        hash_.update(data.bytes)
        return hash_.hexdigest()



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base classes for SQLModel schemas."""

from datetime import datetime
from typing import TYPE_CHECKING, Any, Sequence, TypeVar
from uuid import UUID, uuid4

from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, SQLModel

from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from zenml.models.v2.base.base import BaseResponse

    B = TypeVar("B", bound=BaseResponse)  # type: ignore[type-arg]


class BaseSchema(SQLModel):
    """Base SQL Model for ZenML entities."""

    id: UUID = Field(default_factory=uuid4, primary_key=True)
    created: datetime = Field(default_factory=utc_now)
    updated: datetime = Field(default_factory=utc_now)

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        This method should return query options that improve the performance
        when trying to later on converting that schema to a model.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        return []

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Any:
        """In case the Schema has a corresponding Model, this allows conversion to that model.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic

        Raises:
            NotImplementedError: When the base class fails to implement this.
        """
        raise NotImplementedError(
            "No 'to_model()' method implemented for this"
            f"schema: '{self.__class__.__name__}'."
        )


class NamedSchema(BaseSchema):
    """Base Named SQL Model."""

    name: str



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Thread-safe runtime context for deployments.

This module provides request-scoped state for deployment invocations using
contextvars to ensure thread safety and proper request isolation. Each
deployment request gets its own isolated context that doesn't interfere
with concurrent requests.

It also provides parameter override functionality for the orchestrator
to access deployment parameters without tight coupling.
"""

import contextvars
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from zenml.models import PipelineSnapshotResponse


class _DeploymentState(BaseModel):
    model_config = {"extra": "forbid"}

    active: bool = False
    skip_artifact_materialization: bool = False
    request_id: Optional[str] = None
    snapshot_id: Optional[str] = None
    pipeline_parameters: Dict[str, Any] = Field(default_factory=dict)
    outputs: Dict[str, Dict[str, Any]] = Field(default_factory=dict)

    # In-memory data storage for artifacts
    in_memory_data: Dict[str, Any] = Field(default_factory=dict)

    def reset(self) -> None:
        """Reset the deployment state."""
        self.active = False
        self.request_id = None
        self.snapshot_id = None
        self.pipeline_parameters = {}
        self.outputs = {}
        self.skip_artifact_materialization = False
        self.in_memory_data = {}


_deployment_context: contextvars.ContextVar[_DeploymentState] = (
    contextvars.ContextVar("deployment_context", default=_DeploymentState())
)


def _get_context() -> _DeploymentState:
    """Get the current deployment context state.

    Returns:
        The current deployment context state.
    """
    return _deployment_context.get()


def start(
    request_id: str,
    snapshot: PipelineSnapshotResponse,
    parameters: Dict[str, Any],
    skip_artifact_materialization: bool = False,
) -> None:
    """Initialize deployment state for the current request context.

    Args:
        request_id: The ID of the request.
        snapshot: The snapshot to deploy.
        parameters: The parameters to deploy.
        skip_artifact_materialization: Whether to skip artifact materialization.
    """
    state = _DeploymentState()
    state.active = True
    state.request_id = request_id
    state.snapshot_id = str(snapshot.id)
    state.pipeline_parameters = parameters
    state.outputs = {}
    state.skip_artifact_materialization = skip_artifact_materialization
    _deployment_context.set(state)


def stop() -> None:
    """Clear the deployment state for the current request context."""
    state = _get_context()
    state.reset()


def is_active() -> bool:
    """Return whether deployment state is active in the current context.

    Returns:
        True if the deployment state is active in the current context, False otherwise.
    """
    return _get_context().active


def record_step_outputs(step_name: str, outputs: Dict[str, Any]) -> None:
    """Record raw outputs for a step by invocation id.

    Args:
        step_name: The name of the step to record the outputs for.
        outputs: A dictionary of outputs to record.
    """
    state = _get_context()
    if not state.active:
        return
    if not outputs:
        return
    state.outputs.setdefault(step_name, {}).update(outputs)


def get_outputs() -> Dict[str, Dict[str, Any]]:
    """Return the outputs for all steps in the current context.

    Returns:
        A dictionary of outputs for all steps.
    """
    return dict(_get_context().outputs)


def should_skip_artifact_materialization() -> bool:
    """Check if the current request should skip artifact materialization.

    Returns:
        True if artifact materialization is skipped for this request.
    """
    if is_active():
        state = _get_context()
        return state.skip_artifact_materialization
    return False


def put_in_memory_data(uri: str, data: Any) -> None:
    """Store data in memory for the given URI.

    Args:
        uri: The artifact URI to store data for.
        data: The data to store in memory.
    """
    if is_active():
        state = _get_context()
        state.in_memory_data[uri] = data


def get_in_memory_data(uri: str) -> Any:
    """Get data from memory for the given URI.

    Args:
        uri: The artifact URI to retrieve data for.

    Returns:
        The stored data, or None if not found.
    """
    if is_active():
        state = _get_context()
        return state.in_memory_data[uri]
    return None



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML Pipeline Deployment Entrypoint Configuration."""

from typing import Any, Dict, List, Optional
from uuid import UUID

from zenml.client import Client
from zenml.entrypoints.base_entrypoint_configuration import (
    BaseEntrypointConfiguration,
)
from zenml.integrations.registry import integration_registry
from zenml.logger import get_logger
from zenml.models import DeploymentResponse, PipelineSnapshotResponse
from zenml.utils import uuid_utils

logger = get_logger(__name__)

# Deployment-specific entrypoint options
DEPLOYMENT_ID_OPTION = "deployment_id"


class DeploymentEntrypointConfiguration(BaseEntrypointConfiguration):
    """Entrypoint configuration for ZenML Pipeline Deployment.

    This entrypoint configuration handles the startup and configuration
    of the ZenML pipeline deployment FastAPI application.
    """

    def __init__(self, arguments: List[str]):
        """Initializes the entrypoint configuration.

        Args:
            arguments: Command line arguments to configure this object.
        """
        super().__init__(arguments)
        self._deployment: Optional["DeploymentResponse"] = None

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for the deployment entrypoint.

        Returns:
            Set of required option names
        """
        return {
            DEPLOYMENT_ID_OPTION: True,
        }

    @classmethod
    def get_entrypoint_arguments(cls, **kwargs: Any) -> List[str]:
        """Gets arguments for the deployment entrypoint command.

        Args:
            **kwargs: Keyword arguments containing deployment configuration

        Returns:
            List of command-line arguments

        Raises:
            ValueError: If the deployment ID is not a valid UUID.
        """
        # Get base arguments (snapshot_id, etc.)
        base_args = super().get_entrypoint_arguments(**kwargs)

        deployment_id = kwargs.get(DEPLOYMENT_ID_OPTION)
        if not uuid_utils.is_valid_uuid(deployment_id):
            raise ValueError(
                f"Missing or invalid deployment ID as argument for entrypoint "
                f"configuration. Please make sure to pass a valid UUID to "
                f"`{cls.__name__}.{cls.get_entrypoint_arguments.__name__}"
                f"({DEPLOYMENT_ID_OPTION}=<UUID>)`."
            )

        # Add deployment-specific arguments with defaults
        deployment_args = [
            f"--{DEPLOYMENT_ID_OPTION}",
            str(kwargs.get(DEPLOYMENT_ID_OPTION, "")),
        ]

        return base_args + deployment_args

    @property
    def snapshot(self) -> "PipelineSnapshotResponse":
        """The snapshot configured for this entrypoint configuration.

        Returns:
            The snapshot.

        Raises:
            RuntimeError: If the deployment has no snapshot.
        """
        if self._snapshot is None:
            snapshot = self.deployment.snapshot
            if snapshot is None:
                raise RuntimeError(
                    f"Deployment {self.deployment.id} has no snapshot"
                )
            self._snapshot = snapshot
        return self._snapshot

    @property
    def deployment(self) -> "DeploymentResponse":
        """The deployment configured for this entrypoint configuration.

        Returns:
            The deployment.
        """
        if self._deployment is None:
            self._deployment = self._load_deployment()
        return self._deployment

    def _load_deployment(self) -> "DeploymentResponse":
        """Loads the deployment.

        Returns:
            The deployment.
        """
        deployment_id = UUID(self.entrypoint_args[DEPLOYMENT_ID_OPTION])
        deployment = Client().zen_store.get_deployment(
            deployment_id=deployment_id
        )
        return deployment

    def run(self) -> None:
        """Run the ZenML pipeline deployment application.

        This method starts the FastAPI server with the configured parameters
        and the specified pipeline deployment.
        """
        from zenml.deployers.server.app import BaseDeploymentAppRunner

        # Activate integrations to ensure all components are available
        integration_registry.activate_integrations()

        # Download code if necessary (for remote execution environments)
        self.prepare_code_environment()

        app_runner = BaseDeploymentAppRunner.load_app_runner(self.deployment)
        app_runner.run()



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""AWS App Runner deployer flavor."""

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type

from pydantic import Field

from zenml.deployers.base_deployer import (
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.integrations.aws import (
    AWS_CONNECTOR_TYPE,
    AWS_DEPLOYER_FLAVOR,
    AWS_RESOURCE_TYPE,
)
from zenml.models import ServiceConnectorRequirements

if TYPE_CHECKING:
    from zenml.integrations.aws.deployers import AWSDeployer


class AWSDeployerSettings(BaseDeployerSettings):
    """Settings for the AWS App Runner deployer."""

    region: Optional[str] = Field(
        default=None,
        description="AWS region where the App Runner service will be deployed. "
        "If not specified, the region will be determined from the authenticated "
        "session (service connector or implicit credentials). "
        "App Runner is available in specific regions: "
        "https://docs.aws.amazon.com/apprunner/latest/dg/regions.html",
    )
    service_name_prefix: str = Field(
        default="zenml-",
        description="Prefix for service names in App Runner to avoid naming "
        "conflicts.",
    )

    # Health check configuration
    health_check_grace_period_seconds: int = Field(
        default=20,
        ge=0,
        le=20,
        description="Grace period for health checks in seconds. Range: 0-20.",
    )

    health_check_interval_seconds: int = Field(
        default=10,
        ge=1,
        le=20,
        description="Interval between health checks in seconds. Range: 1-20.",
    )

    health_check_protocol: str = Field(
        default="TCP",
        description="Health check protocol. Options: 'TCP', 'HTTP'.",
        pattern="^TCP|HTTP$",
    )

    health_check_timeout_seconds: int = Field(
        default=2,
        ge=1,
        le=20,
        description="Timeout for health checks in seconds. Range: 1-20.",
    )

    health_check_healthy_threshold: int = Field(
        default=1,
        ge=1,
        le=20,
        description="Number of consecutive successful health checks required.",
    )

    health_check_unhealthy_threshold: int = Field(
        default=5,
        ge=1,
        le=20,
        description="Number of consecutive failed health checks before unhealthy.",
    )

    # Networking and security
    is_publicly_accessible: bool = Field(
        default=True,
        description="Whether the App Runner service is publicly accessible.",
    )

    ingress_vpc_configuration: Optional[str] = Field(
        default=None,
        description="VPC configuration for private App Runner services. "
        "JSON string with VpcId, VpcEndpointId, and VpcIngressConnectionName.",
    )

    # Environment and configuration
    environment_variables: Dict[str, str] = Field(
        default_factory=dict,
        description="Environment variables to set in the App Runner service.",
    )

    # Tags
    tags: Dict[str, str] = Field(
        default_factory=dict,
        description="Tags to apply to the App Runner service.",
    )

    # Secret management configuration
    use_secrets_manager: bool = Field(
        default=True,
        description="Whether to store sensitive environment variables in AWS "
        "Secrets Manager instead of directly in the App Runner service "
        "configuration.",
    )

    secret_name_prefix: str = Field(
        default="zenml-",
        description="Prefix for secret names in Secrets Manager to avoid naming "
        "conflicts.",
    )

    # Observability
    observability_configuration_arn: Optional[str] = Field(
        default=None,
        description="ARN of the observability configuration to associate with "
        "the App Runner service.",
    )

    # Encryption
    encryption_kms_key: Optional[str] = Field(
        default=None,
        description="KMS key ARN for encrypting App Runner service data.",
    )

    # IAM Roles
    instance_role_arn: Optional[str] = Field(
        default=None,
        description="ARN of the IAM role to assign to the App Runner service instances.",
    )

    access_role_arn: Optional[str] = Field(
        default=None,
        description="ARN of the IAM role that App Runner uses to access the "
        "image repository (ECR). Required for private ECR repositories. If not "
        "specified, App Runner will attempt to use the default service role, "
        "which may not have ECR access permissions.",
    )

    # Traffic allocation for A/B testing and gradual rollouts
    traffic_allocation: Dict[str, int] = Field(
        default_factory=lambda: {"LATEST": 100},
        description="Traffic allocation between revisions for A/B testing and "
        "gradual rollouts. Keys can be revision names, tags, or 'LATEST' for "
        "the most recent revision. Values are percentages that must sum to 100. "
        "Example: {'LATEST': 80, 'my-stable-revision': 20}",
    )

    # Resource matching
    strict_resource_matching: bool = Field(
        default=False,
        description="Whether to enforce strict matching of resource requirements "
        "to AWS App Runner supported CPU (vCPU) and memory (GB) combinations. "
        "When True, raises an error if no exact match is found. When False, "
        "automatically selects the closest matching supported combination. "
        "See https://docs.aws.amazon.com/apprunner/latest/dg/architecture.html#architecture.vcpu-memory "
        "for more details.",
    )


# AWS App Runner supported CPU (vCPU) and memory (GB) combinations
DEFAULT_RESOURCE_COMBINATIONS = [
    (
        0.25,
        0.5,
    ),
    (
        0.25,
        1.0,
    ),
    (
        0.5,
        1.0,
    ),
    (
        1.0,
        2.0,
    ),
    (
        1.0,
        3.0,
    ),
    (
        1.0,
        4.0,
    ),
    (
        2.0,
        4.0,
    ),
    (
        2.0,
        6.0,
    ),
    (
        4.0,
        8.0,
    ),
    (
        4.0,
        10.0,
    ),
    (
        4.0,
        12.0,
    ),
]


class AWSDeployerConfig(
    BaseDeployerConfig,
    AWSDeployerSettings,
):
    """Configuration for the AWS App Runner deployer."""

    resource_combinations: List[Tuple[float, float]] = Field(
        default=DEFAULT_RESOURCE_COMBINATIONS,
        description="AWS App Runner supported CPU (vCPU), memory (GB) "
        "combinations.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True


class AWSDeployerFlavor(BaseDeployerFlavor):
    """AWS App Runner deployer flavor."""

    @property
    def name(self) -> str:
        """Name of the deployer flavor.

        Returns:
            Name of the deployer flavor.
        """
        return AWS_DEPLOYER_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the deployer flavor.

        Returns:
            The display name of the deployer flavor.
        """
        return "AWS"

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            connector_type=AWS_CONNECTOR_TYPE,
            resource_type=AWS_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/deployer/aws-app-runner.png"

    @property
    def config_class(self) -> Type[AWSDeployerConfig]:
        """Returns the AWSDeployerConfig config class.

        Returns:
                The config class.
        """
        return AWSDeployerConfig

    @property
    def implementation_class(self) -> Type["AWSDeployer"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        from zenml.integrations.aws.deployers import AWSDeployer

        return AWSDeployer



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utilities for pydantic models."""

import inspect
import json
from json.decoder import JSONDecodeError
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import yaml
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    PlainValidator,
    TypeAdapter,
    ValidationInfo,
    WrapValidator,
    create_model,
    validate_call,
)
from pydantic._internal import _repr as pydantic_repr
from pydantic.fields import FieldInfo
from pydantic.v1.utils import sequence_like

from zenml.logger import get_logger
from zenml.utils import dict_utils, typing_utils, yaml_utils
from zenml.utils.json_utils import pydantic_encoder

logger = get_logger(__name__)

M = TypeVar("M", bound="BaseModel")


def update_model(
    original: M,
    update: Union["BaseModel", Dict[str, Any]],
    recursive: bool = True,
    exclude_none: bool = False,
) -> M:
    """Updates a pydantic model.

    Args:
        original: The model to update.
        update: The update values.
        recursive: If `True`, dictionary values will be updated recursively.
        exclude_none: If `True`, `None` values in the update will be removed.

    Returns:
        The updated model.
    """
    if isinstance(update, Dict):
        if exclude_none:
            update_dict = dict_utils.remove_none_values(
                update, recursive=recursive
            )
        else:
            update_dict = update
    else:
        update_dict = update.model_dump(
            exclude_unset=True, exclude_none=exclude_none
        )

    original_dict = original.model_dump(exclude_unset=True)
    if recursive:
        values = dict_utils.recursive_update(original_dict, update_dict)
    else:
        values = {**original_dict, **update_dict}

    return original.__class__.model_validate(values)


def get_json_schema_for_type(value_type: Any) -> Dict[str, Any]:
    """Get a JSON schema for a Python type using Pydantic.

    Args:
        value_type: Type annotation for which to get the JSON schema.

    Returns:
        JSON schema dictionary.

    Raises:
        ValueError: If the input value is not supported by Pydantic schema
            generation.
    """
    try:
        return TypeAdapter(value_type).json_schema()
    except Exception as e:
        raise ValueError(
            f"Unsupported type `{value_type}`. Only JSON-serializable "
            "types are supported. See the above pydantic error for more "
            "details."
        ) from e


def create_parameter_model(
    model_name: str,
    parameters: Mapping[str, inspect.Parameter],
    default_values: Optional[Mapping[str, Any]] = None,
    config: Optional[ConfigDict] = None,
) -> Type[BaseModel]:
    """Creates a Pydantic model from function parameters.

    Args:
        model_name: Name of the generated model.
        parameters: Mapping of parameter names to function parameters.
        default_values: Optional override defaults keyed by parameter name.
        config: Optional Pydantic config for the generated model.

    Returns:
        A Pydantic model representing the provided parameters.
    """
    fields: Dict[str, Any] = {}
    default_values = default_values or {}

    for name, parameter in parameters.items():
        field_name = name
        if name in default_values:
            default_value = default_values[name]
        elif parameter.default is not inspect.Parameter.empty:
            default_value = parameter.default
        else:
            default_value = ...

        if name.startswith("_"):
            # Pydantic treats field names starting with an underscore as
            # private attributes, so we use a sanitized field name with an
            # alias.
            field_name = name.lstrip("_")
            while field_name in fields:
                field_name = field_name + "_"

            default_value = FieldInfo(
                default=default_value,
                validation_alias=name,
                serialization_alias=name,
            )

        fields[field_name] = (parameter.annotation, default_value)

    model_config = ConfigDict(extra="forbid")
    if config:
        model_config.update(config)
    model_config["serialize_by_alias"] = True

    fields["__config__"] = model_config
    return create_model(model_name, **fields)


class TemplateGenerator:
    """Class to generate templates for pydantic models or classes."""

    def __init__(
        self, instance_or_class: Union[BaseModel, Type[BaseModel]]
    ) -> None:
        """Initializes the template generator.

        Args:
            instance_or_class: The pydantic model or model class for which to
                generate a template.
        """
        self.instance_or_class = instance_or_class

    def run(self) -> Dict[str, Any]:
        """Generates the template.

        Returns:
            The template dictionary.
        """
        if isinstance(self.instance_or_class, BaseModel):
            template = self._generate_template_for_model(
                self.instance_or_class
            )
        else:
            template = self._generate_template_for_model_class(
                self.instance_or_class
            )

        # Convert to json in an intermediate step, so we can leverage Pydantic's
        # encoder to support types like UUID and datetime
        json_string = json.dumps(template, default=pydantic_encoder)
        return cast(Dict[str, Any], json.loads(json_string))

    def _generate_template_for_model(self, model: BaseModel) -> Dict[str, Any]:
        """Generates a template for a pydantic model.

        Args:
            model: The model for which to generate the template.

        Returns:
            The model template.
        """
        template = self._generate_template_for_model_class(model.__class__)

        for name in model.model_fields_set:
            value = getattr(model, name)
            template[name] = self._generate_template_for_value(value)

        return template

    def _generate_template_for_model_class(
        self,
        model_class: Type[BaseModel],
    ) -> Dict[str, Any]:
        """Generates a template for a pydantic model class.

        Args:
            model_class: The model class for which to generate the template.

        Returns:
            The model class template.
        """
        template: Dict[str, Any] = {}

        for name, field in model_class.model_fields.items():
            annotation = field.annotation

            if annotation is not None:
                if self._is_model_class(annotation):
                    template[name] = self._generate_template_for_model_class(
                        annotation
                    )

                elif typing_utils.is_optional(
                    annotation
                ) and self._is_model_class(
                    typing_utils.get_args(annotation)[0]
                ):
                    template[name] = self._generate_template_for_model_class(
                        typing_utils.get_args(annotation)[0]
                    )
                else:
                    template[name] = pydantic_repr.display_as_type(annotation)

        return template

    def _generate_template_for_value(self, value: Any) -> Any:
        """Generates a template for an arbitrary value.

        Args:
            value: The value for which to generate the template.

        Returns:
            The value template.
        """
        if isinstance(value, Dict):
            return {
                k: self._generate_template_for_value(v)
                for k, v in value.items()
            }
        elif sequence_like(value):
            return [self._generate_template_for_value(v) for v in value]
        elif isinstance(value, BaseModel):
            return self._generate_template_for_model(value)
        else:
            return value

    @staticmethod
    def _is_model_class(value: Any) -> bool:
        """Checks if the given value is a pydantic model class.

        Args:
            value: The value to check.

        Returns:
            If the value is a pydantic model class.
        """
        return isinstance(value, type) and issubclass(value, BaseModel)


class YAMLSerializationMixin(BaseModel):
    """Class to serialize/deserialize pydantic models to/from YAML."""

    def yaml(self, sort_keys: bool = False, **kwargs: Any) -> str:
        """YAML string representation..

        Args:
            sort_keys: Whether to sort the keys in the YAML representation.
            **kwargs: Kwargs to pass to the pydantic model_dump(...) method.

        Returns:
            YAML string representation.
        """
        dict_ = json.loads(
            json.dumps(
                self.model_dump(mode="json", **kwargs), sort_keys=sort_keys
            )
        )
        return yaml.dump(dict_, sort_keys=sort_keys)

    @classmethod
    def from_yaml(cls: Type[M], path: str) -> M:
        """Creates an instance from a YAML file.

        Args:
            path: Path to a YAML file.

        Returns:
            The model instance.
        """
        dict_ = yaml_utils.read_yaml(path)
        return cls.model_validate(dict_)


def validate_function_args(
    __func: Callable[..., Any],
    __config: Optional[ConfigDict],
    *args: Any,
    **kwargs: Any,
) -> Dict[str, Any]:
    """Validates arguments passed to a function.

    This function validates that all arguments to call the function exist and
    that the types match.

    It raises a pydantic.ValidationError if the validation fails.

    Args:
        __func: The function for which the arguments are passed.
        __config: The pydantic config for the underlying model that is created
            to validate the types of the arguments.
        *args: Function arguments.
        **kwargs: Function keyword arguments.

    Returns:
        The validated arguments.
    """
    signature = inspect.signature(__func)

    validated_args = ()
    validated_kwargs = {}

    def f(*args: Any, **kwargs: Dict[Any, Any]) -> None:
        nonlocal validated_args
        nonlocal validated_kwargs

        validated_args = args
        validated_kwargs = kwargs

    # We create a dummy function with the original function signature to run
    # pydantic validation without actually running the function code
    f.__signature__ = signature  # type: ignore[attr-defined]
    f.__annotations__ = __func.__annotations__

    validated_function = validate_call(config=__config, validate_return=False)(
        f
    )

    # This raises a pydantic.ValidatonError in case the arguments are not valid
    validated_function(*args, **kwargs)

    bound_arguments = dict(
        signature.bind(*validated_args, **validated_kwargs).arguments
    )

    variadic_keyword_parameter = next(
        (
            parameter.name
            for parameter in signature.parameters.values()
            if parameter.kind is inspect.Parameter.VAR_KEYWORD
        ),
        None,
    )
    if variadic_keyword_parameter and isinstance(
        bound_arguments.get(variadic_keyword_parameter), dict
    ):
        # If the function defines a variadic keyword parameter, we expand the
        # and include them flat in the return dictionary.
        variadic_keyword_arguments = bound_arguments.pop(
            variadic_keyword_parameter
        )
        bound_arguments.update(variadic_keyword_arguments)

    return bound_arguments


def model_validator_data_handler(
    raw_data: Any,
    base_class: Type[BaseModel],
    validation_info: ValidationInfo,
) -> Dict[str, Any]:
    """Utility function to parse raw input data of varying types to a dict.

    With the change to pydantic v2, validators which operate with "before"
    (or previously known as the "pre" parameter) are getting "Any" types of raw
    input instead of a "Dict[str, Any]" as before. Depending on the use-case,
    this can create conflicts after the migration and this function will be
    used as a helper function to handle different types of raw input data.

    A code snippet to showcase how the behavior changes. The "before" validator
    prints the type of the input:

        class Base(BaseModel):
            a: int = 3

        class MyClass(Base):
            @model_validator(mode="before")
            @classmethod
            def before_validator(cls, data: Any) -> Any:
                print(type(data))
                return {}

        one = MyClass() # prints "<class 'dict'>"
        MyClass.model_validate(one)  # prints NOTHING, it is already validated
        MyClass.model_validate("asdf")  # prints "<class 'str'>", fails without the modified return.
        MyClass.model_validate(RandomClass())  # prints "<class 'RandomClass'>", fails without the modified return.
        MyClass.model_validate(Base())  # prints "<class 'Base'>", fails without the modified return.
        MyClass.model_validate_json(json.dumps("aria"))  # prints "<class 'str'>", fails without the modified return.
        MyClass.model_validate_json(json.dumps([1]))  # prints "<class 'list'>", fails without the modified return.
        MyClass.model_validate_json(one.model_dump_json())  # prints "<class 'dict'>"

    Args:
        raw_data: The raw data passed to the validator, can be "Any" type.
        base_class: The class that the validator belongs to
        validation_info: Extra information about the validation process.

    Raises:
        TypeError: if the type of the data is not processable.
        ValueError: in case of an unknown validation mode.

    Returns:
        A dictionary which will be passed to the eventual validator of pydantic.
    """
    if validation_info.mode == "python":
        # This is mode is only active if people validate objects using pythonic
        # raw data such as MyClass(...) or MyClass.model_validate()

        if isinstance(raw_data, dict):
            # In most cases, this is the behavior as the raw input is a dict
            return raw_data

        elif isinstance(raw_data, base_class):
            # In some cases, we pass the same object type to the validation
            # in such cases, it is critical we keep the original structure of
            # fields that are already set.
            return dict(raw_data)

        elif issubclass(base_class, raw_data.__class__):
            # There are a few occurrences where the annotation of the field is
            # denoted by a subclass, and we use the instance of its super class
            # as the raw input. In such cases we will use the same approach as
            # before, while raising a debug message.
            logger.debug(
                f"During the validation of a `{base_class}` object, an instance"
                f"of `{raw_data.__class__}` (super class of `{base_class}`) "
                f"has been passed as raw input. This might lead to unexpected "
                f"behavior in case `{base_class}` have features which can not"
                f"be extracted from an instance of a `{raw_data.__class__}`."
            )
            return dict(raw_data)

        elif isinstance(raw_data, str):
            # If the raw input is a raw string, we can try to use the `json`
            # module to parse it. The resulting data needs to be a proper
            # dict for us to pass it to the validation process.
            try:
                json_data = json.loads(raw_data)

                if isinstance(json_data, dict):
                    return json_data
                else:
                    raise TypeError("The resulting json data is not a dict!")

            except (TypeError, JSONDecodeError) as e:
                raise TypeError(
                    "The raw json input string can not be converted to a "
                    f"dict: {e}"
                )
        else:
            raise TypeError(
                "Unsupported type of raw input data for the `python` validation"
                "mode of the pydantic class. Please consider changing the way "
                f"you are creating using the `{base_class}` or instead use"
                f"`{base_class}.model_validate_json()`."
            )

    elif validation_info.mode == "json":
        # This is mode is only active if people validate objects using json
        # input data such as MyClass.model_validate_json()
        if isinstance(raw_data, dict):
            return raw_data
        else:
            raise TypeError(
                f"The resulting JSON data {raw_data} is not a dict, therefore"
                f"can not be used by the validation process."
            )
    else:
        # Unknown validation mode
        raise ValueError(f"Unknown validation mode. {validation_info.mode}")


def before_validator_handler(
    method: Callable[..., Any],
) -> Callable[[Any, Any, Any], Any]:
    """Decorator to handle the raw input data for pydantic model validators.

    Args:
        method: the class method with the actual validation logic.

    Returns:
        the validator method
    """

    def before_validator(
        cls: Type[BaseModel], data: Any, validation_info: ValidationInfo
    ) -> Any:
        """Wrapper method to handle the raw data.

        Args:
            cls: the class handler
            data: the raw input data
            validation_info: the context of the validation.

        Returns:
            the validated data
        """
        data = model_validator_data_handler(
            raw_data=data, base_class=cls, validation_info=validation_info
        )
        return method(cls=cls, data=data)

    return before_validator


def has_validators(
    pydantic_class: Type[BaseModel],
    field_name: Optional[str] = None,
) -> bool:
    """Function to check if a Pydantic model or a pydantic field has validators.

    Args:
        pydantic_class: The class defining the pydantic model.
        field_name: Optional, field info. If specified, this function will focus
            on a singular field within the class. If not specified, it will
            check model validators.

    Returns:
        Whether the specified field or class has a validator or not.
    """
    # If field is not specified check model validators
    if field_name is None:
        if pydantic_class.__pydantic_decorators__.model_validators:
            return True

    # Else, check field validators
    else:
        # 1. Field validators can be defined through @field_validator decorators
        f_validators = pydantic_class.__pydantic_decorators__.field_validators

        for name, f_v in f_validators.items():
            if field_name in f_v.info.fields:
                return True

        # 2. Field validators can be defined through the Annotation[.....]
        field_info = pydantic_class.model_fields[field_name]
        if metadata := field_info.metadata:
            if any(
                isinstance(
                    m,
                    (
                        AfterValidator,
                        BeforeValidator,
                        PlainValidator,
                        WrapValidator,
                    ),
                )
                for m in metadata
            ):
                return True

    return False



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:

#       https://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality to deploy a ZenML stack to AWS."""

from typing import ClassVar, Dict, List, Optional

from zenml.enums import StackDeploymentProvider
from zenml.models import StackDeploymentConfig
from zenml.stack_deployments.constants import (
    TERRAFORM_AWS_MODULE_VERSION_SPEC,
    TERRAFORM_PROVIDER_VERSION_SPEC,
)
from zenml.stack_deployments.stack_deployment import (
    STACK_DEPLOYMENT_TERRAFORM,
    ZenMLCloudStackDeployment,
)
from zenml.utils.string_utils import random_str

AWS_DEPLOYMENT_TYPE = "cloud-formation"


class AWSZenMLCloudStackDeployment(ZenMLCloudStackDeployment):
    """AWS ZenML Cloud Stack Deployment."""

    provider: ClassVar[StackDeploymentProvider] = StackDeploymentProvider.AWS
    deployment: ClassVar[str] = AWS_DEPLOYMENT_TYPE

    @classmethod
    def description(cls) -> str:
        """Return a description of the ZenML Cloud Stack Deployment.

        This will be displayed when the user is prompted to deploy
        the ZenML stack.

        Returns:
            A MarkDown description of the ZenML Cloud Stack Deployment.
        """
        return """
Provision and register a basic AWS ZenML stack authenticated and connected to
all the necessary cloud infrastructure resources required to run pipelines in
AWS.
"""

    @classmethod
    def instructions(cls) -> str:
        """Return instructions on how to deploy the ZenML stack to the specified cloud provider.

        This will be displayed before the user is prompted to deploy the ZenML
        stack.

        Returns:
            MarkDown instructions on how to deploy the ZenML stack to the
            specified cloud provider.
        """
        return """
You will be redirected to the AWS console in your browser where you'll be asked
to log into your AWS account and create a CloudFormation ZenML stack. The stack
parameters will be pre-filled with the necessary information to connect ZenML to
your AWS account, so you should only need to review and confirm the stack.

**NOTE**: The CloudFormation stack will create the following new resources in
your AWS account. Please ensure you have the necessary permissions and are aware
of any potential costs:

- An S3 bucket registered as a [ZenML artifact store](https://docs.zenml.io/stack-components/artifact-stores/s3).
- An ECR repository registered as a [ZenML container registry](https://docs.zenml.io/stack-components/container-registries/aws).
- Sagemaker registered as a [ZenML orchestrator](https://docs.zenml.io/stack-components/orchestrators/sagemaker)
as well as a [ZenML step operator](https://docs.zenml.io/stack-components/step-operators/sagemaker).
- A CodeBuild project registered as a [ZenML image builder](https://docs.zenml.io/stack-components/image-builder/aws).
- An IAM user and IAM role with the minimum necessary permissions to access the
above resources.
- An AWS access key used to give access to ZenML to connect to the above
resources through a [ZenML service connector](https://docs.zenml.io/how-to/auth-management/aws-service-connector).

The CloudFormation stack will automatically create an AWS secret key and
will share it with ZenML to give it permission to access the resources created
by the stack. You can revoke these permissions at any time by deleting the
CloudFormation stack.

**Estimated costs**

A small training job would cost around: $0.60

These are rough estimates and actual costs may vary based on your usage and specific AWS pricing. 
Some services may be eligible for the AWS Free Tier. Use [the AWS Pricing Calculator](https://calculator.aws)
for a detailed estimate based on your usage.

💡 **After the CloudFormation stack is deployed, you can return to the CLI to
view details about the associated ZenML stack automatically registered with
ZenML.**
"""

    @classmethod
    def post_deploy_instructions(cls) -> str:
        """Return instructions on what to do after the deployment is complete.

        This will be displayed after the deployment is complete.

        Returns:
            MarkDown instructions on what to do after the deployment is
            complete.
        """
        return """
The ZenML stack has been successfully deployed and registered. You can delete
the CloudFormation at any time to revoke ZenML's access to your AWS account and
to clean up the resources created by the stack by using the AWS CloudFormation
console.
"""

    @classmethod
    def integrations(cls) -> List[str]:
        """Return the ZenML integrations required for the stack.

        Returns:
            The list of ZenML integrations that need to be installed for the
            stack to be usable.
        """
        return [
            "aws",
            "s3",
        ]

    @classmethod
    def permissions(cls) -> Dict[str, List[str]]:
        """Return the permissions granted to ZenML to access the cloud resources.

        Returns:
            The permissions granted to ZenML to access the cloud resources, as
            a dictionary grouping permissions by resource.
        """
        return {
            "S3 Bucket": [
                "s3:ListBucket",
                "s3:GetObject",
                "s3:PutObject",
                "s3:DeleteObject",
                "s3:GetBucketVersioning",
                "s3:ListBucketVersions",
                "s3:DeleteObjectVersion",
            ],
            "ECR Repository": [
                "ecr:DescribeRepositories",
                "ecr:ListRepositories",
                "ecr:DescribeRegistry",
                "ecr:BatchGetImage",
                "ecr:DescribeImages",
                "ecr:BatchCheckLayerAvailability",
                "ecr:GetDownloadUrlForLayer",
                "ecr:InitiateLayerUpload",
                "ecr:UploadLayerPart",
                "ecr:CompleteLayerUpload",
                "ecr:PutImage",
                "ecr:GetAuthorizationToken",
            ],
            "CloudBuild (Client)": [
                "codebuild:CreateProject",
                "codebuild:BatchGetBuilds",
            ],
            "CloudBuild (Service)": [
                "s3:GetObject",
                "s3:GetObjectVersion",
                "logs:CreateLogGroup",
                "logs:CreateLogStream",
                "logs:PutLogEvents",
                "ecr:BatchGetImage",
                "ecr:DescribeImages",
                "ecr:BatchCheckLayerAvailability",
                "ecr:GetDownloadUrlForLayer",
                "ecr:InitiateLayerUpload",
                "ecr:UploadLayerPart",
                "ecr:CompleteLayerUpload",
                "ecr:PutImage",
                "ecr:GetAuthorizationToken",
            ],
            "SageMaker (Client)": [
                "sagemaker:CreatePipeline",
                "sagemaker:StartPipelineExecution",
                "sagemaker:DescribePipeline",
                "sagemaker:DescribePipelineExecution",
            ],
            "SageMaker (Jobs)": [
                "AmazonSageMakerFullAccess",
            ],
        }

    @classmethod
    def locations(cls) -> Dict[str, str]:
        """Return the locations where the ZenML stack can be deployed.

        Returns:
            The regions where the ZenML stack can be deployed as a map of region
            names to region descriptions.
        """
        # Return a list of all possible AWS regions

        # Based on the AWS regions listed at
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/using-regions-availability-zones.html
        return {
            "Africa (Cape Town)": "af-south-1",
            "Asia Pacific (Hong Kong)": "ap-east-1",
            "Asia Pacific (Hyderabad)": "ap-south-2",
            "Asia Pacific (Jakarta)": "ap-southeast-3",
            "Asia Pacific (Melbourne)": "ap-southeast-4",
            "Asia Pacific (Mumbai)": "ap-south-1",
            "Asia Pacific (Osaka)": "ap-northeast-3",
            "Asia Pacific (Seoul)": "ap-northeast-2",
            "Asia Pacific (Singapore)": "ap-southeast-1",
            "Asia Pacific (Sydney)": "ap-southeast-2",
            "Asia Pacific (Tokyo)": "ap-northeast-1",
            "Canada (Central)": "ca-central-1",
            "Canada West (Calgary)": "ca-west-1",
            "Europe (Frankfurt)": "eu-central-1",
            "Europe (Ireland)": "eu-west-1",
            "Europe (London)": "eu-west-2",
            "Europe (Milan)": "eu-south-1",
            "Europe (Paris)": "eu-west-3",
            "Europe (Spain)": "eu-south-2",
            "Europe (Stockholm)": "eu-north-1",
            "Europe (Zurich)": "eu-central-2",
            "Israel (Tel Aviv)": "il-central-1",
            "Middle East (Bahrain)": "me-south-1",
            "Middle East (UAE)": "me-central-1",
            "South America (São Paulo)": "sa-east-1",
            "US East (Ohio)": "us-east-2",
            "US East (N. Virginia)": "us-east-1",
            "US West (N. California)": "us-west-1",
            "US West (Oregon)": "us-west-2",
        }

    def get_deployment_config(
        self,
    ) -> StackDeploymentConfig:
        """Return the configuration to deploy the ZenML stack to the specified cloud provider.

        The configuration should include:

        * a cloud provider console URL where the user will be redirected to
        deploy the ZenML stack. The URL should include as many pre-filled
        URL query parameters as possible.
        * a textual description of the URL
        * a Terraform script used to deploy the ZenML stack
        * some deployment providers may require additional configuration
        parameters or scripts to be passed to the cloud provider in addition to
        the deployment URL query parameters. Where that is the case, this method
        should also return a string that the user can copy and paste into the
        cloud provider console to deploy the ZenML stack (e.g. a set of
        environment variables, YAML configuration snippet, bash or Terraform
        script etc.).

        Returns:
            The configuration or script to deploy the ZenML stack to the
            specified cloud provider.
        """
        params = dict(
            stackName=self.stack_name,
            templateURL="https://zenml-cf-templates.s3.eu-central-1.amazonaws.com/aws-ecr-s3-sagemaker.yaml",
            param_ResourceName=f"zenml-{random_str(6).lower()}",
            param_ZenMLServerURL=self.zenml_server_url,
            param_ZenMLServerAPIToken=self.zenml_server_api_token,
            param_CodeBuild="true",
        )
        # Encode the parameters as URL query parameters
        query_params = "&".join([f"{k}={v}" for k, v in params.items()])

        region = ""
        if self.location:
            region = f"region={self.location}"

        url = (
            f"https://console.aws.amazon.com/cloudformation/home?"
            f"{region}#/stacks/create/review?{query_params}"
        )

        config: Optional[str] = None
        if self.deployment_type == STACK_DEPLOYMENT_TERRAFORM:
            config = f"""terraform {{
    required_providers {{
        aws = {{
            source  = "hashicorp/aws"
        }}
        zenml = {{
            source = "zenml-io/zenml"
            version = "{TERRAFORM_PROVIDER_VERSION_SPEC}"
        }}
    }}
}}

provider "aws" {{
    region = "{self.location or "eu-central-1"}"
}}

provider "zenml" {{
    server_url = "{self.zenml_server_url}"
    api_token = "{self.zenml_server_api_token}"
}}

module "zenml_stack" {{
    source  = "zenml-io/zenml-stack/aws"
    version = "{TERRAFORM_AWS_MODULE_VERSION_SPEC}"
    zenml_stack_name = "{self.stack_name}"
    zenml_stack_deployment = "{self.deployment_type}"
}}
output "zenml_stack_id" {{
    value = module.zenml_stack.zenml_stack_id
}}
output "zenml_stack_name" {{
    value = module.zenml_stack.zenml_stack_name
}}"""

        return StackDeploymentConfig(
            deployment_url=url,
            deployment_url_text="AWS CloudFormation Console",
            configuration=config,
        )



"""Release [0.93.1].

Revision ID: 0.93.1
Revises: 40a700c16141
Create Date: 2026-01-13 13:18:35.733568

"""

# revision identifiers, used by Alembic.
revision = "0.93.1"
down_revision = "40a700c16141"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



"""add snapshot index [d1e8c2c4a9ef].

Revision ID: d1e8c2c4a9ef
Revises: 0e5c99d02aa7
Create Date: 2026-01-23 16:25:05.328420

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "d1e8c2c4a9ef"
down_revision = "0e5c99d02aa7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.create_index(
            "ix_pipeline_snapshot_source_snapshot_id",
            ["source_snapshot_id"],
            unique=False,
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.drop_index("ix_pipeline_snapshot_source_snapshot_id")



"""remove db dependency loop [2d201872e23c].

Revision ID: 2d201872e23c
Revises: 0.56.3
Create Date: 2024-04-11 09:47:47.557220

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "2d201872e23c"
down_revision = "0.56.3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_pipeline_build_template_deployment_id_pipeline_deployment",
            type_="foreignkey",
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQL Model Implementations for code repositories."""

import json
from typing import Any, Optional, Sequence
from uuid import UUID

from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship

from zenml.models import (
    CodeReferenceRequest,
    CodeReferenceResponse,
    CodeReferenceResponseBody,
    CodeReferenceResponseMetadata,
    CodeRepositoryRequest,
    CodeRepositoryResponse,
    CodeRepositoryResponseBody,
    CodeRepositoryResponseMetadata,
    CodeRepositoryResponseResources,
    CodeRepositoryUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import BaseSchema, NamedSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg


class CodeRepositorySchema(NamedSchema, table=True):
    """SQL Model for code repositories."""

    __tablename__ = "code_repository"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_code_repository_name_in_project",
        ),
    )

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    project: "ProjectSchema" = Relationship(back_populates="code_repositories")

    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    user: Optional["UserSchema"] = Relationship(
        back_populates="code_repositories"
    )

    config: str = Field(sa_column=Column(TEXT, nullable=False))
    source: str = Field(sa_column=Column(TEXT, nullable=False))
    logo_url: Optional[str] = Field()
    description: Optional[str] = Field(sa_column=Column(TEXT, nullable=True))

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(CodeRepositorySchema.user)),
                ]
            )

        return options

    @classmethod
    def from_request(
        cls, request: "CodeRepositoryRequest"
    ) -> "CodeRepositorySchema":
        """Convert a `CodeRepositoryRequest` to a `CodeRepositorySchema`.

        Args:
            request: The request model to convert.

        Returns:
            The converted schema.
        """
        return cls(
            name=request.name,
            project_id=request.project,
            user_id=request.user,
            config=json.dumps(request.config),
            source=request.source.model_dump_json(),
            description=request.description,
            logo_url=request.logo_url,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> "CodeRepositoryResponse":
        """Convert a `CodeRepositorySchema` to a `CodeRepositoryResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The created CodeRepositoryResponse.
        """
        body = CodeRepositoryResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            source=json.loads(self.source),
            logo_url=self.logo_url,
            created=self.created,
            updated=self.updated,
        )
        metadata = None
        if include_metadata:
            metadata = CodeRepositoryResponseMetadata(
                config=json.loads(self.config),
                description=self.description,
            )

        resources = None
        if include_resources:
            resources = CodeRepositoryResponseResources(
                user=self.user.to_model() if self.user else None,
            )

        return CodeRepositoryResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(self, update: "CodeRepositoryUpdate") -> "CodeRepositorySchema":
        """Update a `CodeRepositorySchema` with a `CodeRepositoryUpdate`.

        Args:
            update: The update model.

        Returns:
            The updated `CodeRepositorySchema`.
        """
        if update.name:
            self.name = update.name

        if update.description:
            self.description = update.description

        if update.logo_url:
            self.logo_url = update.logo_url

        if update.config:
            self.config = json.dumps(update.config)

        self.updated = utc_now()
        return self


class CodeReferenceSchema(BaseSchema, table=True):
    """SQL Model for code references."""

    __tablename__ = "code_reference"

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    project: "ProjectSchema" = Relationship()

    code_repository_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=CodeRepositorySchema.__tablename__,
        source_column="code_repository_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    code_repository: "CodeRepositorySchema" = Relationship()

    commit: str
    subdirectory: str

    @classmethod
    def from_request(
        cls, request: "CodeReferenceRequest", project_id: UUID
    ) -> "CodeReferenceSchema":
        """Convert a `CodeReferenceRequest` to a `CodeReferenceSchema`.

        Args:
            request: The request model to convert.
            project_id: The project ID.

        Returns:
            The converted schema.
        """
        return cls(
            project_id=project_id,
            commit=request.commit,
            subdirectory=request.subdirectory,
            code_repository_id=request.code_repository,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> "CodeReferenceResponse":
        """Convert a `CodeReferenceSchema` to a `CodeReferenceResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic

            kwargs: Additional keyword arguments.

        Returns:
            The converted model.
        """
        body = CodeReferenceResponseBody(
            commit=self.commit,
            subdirectory=self.subdirectory,
            code_repository=self.code_repository.to_model(),
            created=self.created,
            updated=self.updated,
        )
        metadata = None
        if include_metadata:
            metadata = CodeReferenceResponseMetadata()

        return CodeReferenceResponse(
            id=self.id,
            body=body,
            metadata=metadata,
        )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality for OpenAI standard hooks."""

import io
import sys
from typing import Optional

from openai import OpenAI
from rich.console import Console

from zenml import get_step_context
from zenml.client import Client
from zenml.logger import get_logger

logger = get_logger(__name__)


def openai_alerter_failure_hook_helper(
    exception: BaseException,
    model_name: str,
) -> None:
    """Standard failure hook that sends a message to an Alerter.

    Your OpenAI API key must be stored in the secret store under the name
    "openai" and with the key "api_key".

    Args:
        exception: The exception that was raised.
        model_name: The OpenAI model to use for the chatbot.

    This implementation uses the OpenAI v1 SDK with automatic retries and backoff.
    """
    client = Client()
    context = get_step_context()

    # get the api_key from the secret store
    try:
        openai_secret = client.get_secret(
            "openai", allow_partial_name_match=False
        )
        openai_api_key: Optional[str] = openai_secret.secret_values.get(
            "api_key"
        )
    except (KeyError, NotImplementedError):
        openai_api_key = None

    alerter = client.active_stack.alerter
    if alerter and openai_api_key:
        # Capture rich traceback
        output_captured = io.StringIO()
        original_stdout = sys.stdout
        sys.stdout = output_captured
        console = Console()
        console.print_exception(show_locals=False)

        sys.stdout = original_stdout
        rich_traceback = output_captured.getvalue()

        # Initialize OpenAI client with timeout and retry settings
        openai_client = OpenAI(
            api_key=openai_api_key,
            max_retries=3,  # Will retry 3 times with exponential backoff
            timeout=60.0,  # 60 second timeout
        )

        # Create chat completion using the new client pattern
        response = openai_client.chat.completions.create(
            model=model_name,
            messages=[
                {
                    "role": "user",
                    "content": f"This is an error message (following an exception of type '{type(exception)}') "
                    f"I encountered while executing a ZenML step. Please suggest ways I might fix the problem. "
                    f"Feel free to give code snippets as examples, and note that your response will be piped "
                    f"to a Slack bot so make sure the formatting is appropriate: {exception} -- {rich_traceback}. "
                    f"Thank you!",
                }
            ],
        )

        suggestion = response.choices[0].message.content

        # Format the alert message
        message = "\n".join(
            [
                "*Failure Hook Notification! Step failed!*",
                "",
                f"Run name: `{context.pipeline_run.name}`",
                f"Step name: `{context.step_run.name}`",
                f"Parameters: `{context.step_run.config.parameters}`",
                f"Exception: `({type(exception)}) {exception}`",
                "",
                f"*OpenAI ChatGPT's suggestion (model = `{model_name}`) on how to fix it:*\n `{suggestion}`",
            ]
        )

        alerter.post(message)
    elif not openai_api_key:
        logger.warning(
            "Specified OpenAI failure hook but no OpenAI API key found. Skipping..."
        )
    else:
        logger.warning(
            "Specified OpenAI failure hook but no alerter configured in the stack. Skipping..."
        )


def openai_chatgpt_alerter_failure_hook(
    exception: BaseException,
) -> None:
    """Alerter hook that uses the OpenAI ChatGPT model.

    Args:
        exception: The exception that was raised.
    """
    openai_alerter_failure_hook_helper(exception, "gpt-3.5-turbo")


def openai_gpt4_alerter_failure_hook(
    exception: BaseException,
) -> None:
    """Alerter hook that uses the OpenAI GPT-4 model.

    Args:
        exception: The exception that was raised.
    """
    openai_alerter_failure_hook_helper(exception, "gpt-4o")



"""Release [0.61.0].

Revision ID: 0.61.0
Revises: 0d707865f404
Create Date: 2024-07-08 15:52:54.765307

"""

# revision identifiers, used by Alembic.
revision = "0.61.0"
down_revision = "0d707865f404"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation for the Comet experiment tracker."""

import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast

from comet_ml import Experiment  # type: ignore

from zenml.constants import METADATA_EXPERIMENT_TRACKER_URL
from zenml.experiment_trackers.base_experiment_tracker import (
    BaseExperimentTracker,
)
from zenml.integrations.comet.flavors.comet_experiment_tracker_flavor import (
    CometExperimentTrackerConfig,
    CometExperimentTrackerSettings,
)
from zenml.logger import get_logger
from zenml.metadata.metadata_types import Uri

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo
    from zenml.metadata.metadata_types import MetadataType


logger = get_logger(__name__)

COMET_API_KEY = "COMET_API_KEY"


class CometExperimentTracker(BaseExperimentTracker):
    """Track experiment using Comet."""

    @property
    def config(self) -> CometExperimentTrackerConfig:
        """Returns the `CometExperimentTrackerConfig` config.

        Returns:
            The configuration.
        """
        return cast(CometExperimentTrackerConfig, self._config)

    @property
    def settings_class(self) -> Type[CometExperimentTrackerSettings]:
        """Settings class for the Comet experiment tracker.

        Returns:
            The settings class.
        """
        return CometExperimentTrackerSettings

    def prepare_step_run(self, info: "StepRunInfo") -> None:
        """Configures a Comet experiment.

        Args:
            info: Info about the step that will be executed.
        """
        os.environ[COMET_API_KEY] = self.config.api_key
        settings = cast(
            CometExperimentTrackerSettings, self.get_settings(info)
        )
        tags = settings.tags + [info.run_name, info.pipeline.name]
        comet_exp_name = (
            settings.run_name or f"{info.run_name}_{info.pipeline_step_name}"
        )
        self._initialize_comet(
            run_name=comet_exp_name, tags=tags, settings=settings.settings
        )

    def get_step_run_metadata(
        self, info: "StepRunInfo"
    ) -> Dict[str, "MetadataType"]:
        """Get component- and step-specific metadata after a step ran.

        Args:
            info: Info about the step that was executed.

        Returns:
            A dictionary of metadata.
        """
        exp_url: Optional[str] = None
        exp_name: Optional[str] = None

        if self.experiment:
            exp_url = self.experiment.url
            exp_name = self.experiment.name

        # If the URL cannot be retrieved, use the default experiment URL
        default_exp_url = (
            f"https://www.comet.com/{self.config.workspace}/"
            f"{self.config.project_name}/experiments/"
        )
        exp_url = exp_url or default_exp_url

        # If the experiment name cannot be retrieved, use the default name
        default_exp_name = f"{info.run_name}_{info.pipeline_step_name}"
        settings = cast(
            CometExperimentTrackerSettings, self.get_settings(info)
        )
        exp_name = exp_name or settings.run_name or default_exp_name

        return {
            METADATA_EXPERIMENT_TRACKER_URL: Uri(exp_url),
            "comet_experiment_name": exp_name,
        }

    def cleanup_step_run(self, info: "StepRunInfo", step_failed: bool) -> None:
        """Stops the Comet experiment.

        Args:
            info: Info about the step that was executed.
            step_failed: Whether the step failed or not.
        """
        if self.experiment:
            self.experiment.end()
        os.environ.pop(COMET_API_KEY, None)

    def log_metrics(
        self,
        metrics: Dict[str, Any],
        step: Optional[int] = None,
    ) -> None:
        """Logs metrics to the Comet experiment.

        Args:
            metrics: Dictionary of metrics to log.
            step: Optional step number associated with the metrics.
        """
        if self.experiment:
            self.experiment.log_metrics(metrics, step=step)

    def log_params(self, params: Dict[str, Any]) -> None:
        """Logs parameters to the Comet experiment.

        Args:
            params: Dictionary of parameters to log.
        """
        if self.experiment:
            self.experiment.log_parameters(params)

    def _initialize_comet(
        self,
        run_name: str,
        tags: List[str],
        settings: Union[Dict[str, Any], None] = None,
    ) -> None:
        """Initializes a Comet experiment.

        Args:
            run_name: Name of the Comet experiment.
            tags: Tags to attach to the Comet experiment.
            settings: Additional settings for the Comet experiment.
        """
        logger.info(
            f"Initializing Comet with workspace {self.config.workspace}, project "
            f"name: {self.config.project_name}, run_name: {run_name}."
        )
        self.experiment = Experiment(
            workspace=self.config.workspace,
            project_name=self.config.project_name,
            **settings or {},
        )
        self.experiment.set_name(run_name)
        self.experiment.add_tags(tags)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Great Expectations data profiling standard step."""

from typing import Any, Dict, Optional

import pandas as pd
from great_expectations.core import (  # type: ignore[import-untyped]
    ExpectationSuite,
)

from zenml import step
from zenml.integrations.great_expectations.data_validators.ge_data_validator import (
    GreatExpectationsDataValidator,
)


@step
def great_expectations_profiler_step(
    dataset: pd.DataFrame,
    expectation_suite_name: str,
    data_asset_name: Optional[str] = None,
    profiler_kwargs: Optional[Dict[str, Any]] = None,
    overwrite_existing_suite: bool = True,
) -> ExpectationSuite:
    """Infer data validation rules from a pandas dataset.

    Args:
        dataset: The dataset from which the expectation suite will be inferred.
        expectation_suite_name: The name of the expectation suite to infer.
        data_asset_name: The name of the data asset to profile.
        profiler_kwargs: A dictionary of keyword arguments to pass to the
            profiler.
        overwrite_existing_suite: Whether to overwrite an existing expectation
            suite.

    Returns:
        The generated Great Expectations suite.
    """
    data_validator = GreatExpectationsDataValidator.get_active_data_validator()

    return data_validator.data_profiling(
        dataset,
        expectation_suite_name=expectation_suite_name,
        data_asset_name=data_asset_name,
        profiler_kwargs=profiler_kwargs or {},
        overwrite_existing_suite=overwrite_existing_suite,
    )



"""Release [0.56.3].

Revision ID: 0.56.3
Revises: 1d74e596abb8
Create Date: 2024-04-09 14:16:22.179577

"""

# revision identifiers, used by Alembic.
revision = "0.56.3"
down_revision = "1d74e596abb8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base orchestrator class."""

import time
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

from pydantic import model_validator

from zenml.client import Client
from zenml.constants import (
    ENV_ZENML_PREVENT_CLIENT_SIDE_CACHING,
    handle_bool_env_var,
)
from zenml.enums import ExecutionMode, ExecutionStatus, StackComponentType
from zenml.exceptions import (
    HookExecutionException,
    IllegalOperationError,
    RunMonitoringError,
)
from zenml.hooks.hook_validators import load_and_run_hook
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType
from zenml.orchestrators.exceptions import PipelineSubmissionError
from zenml.orchestrators.publish_utils import (
    publish_pipeline_run_metadata,
    publish_pipeline_run_status_update,
    publish_schedule_metadata,
)
from zenml.orchestrators.utils import get_config_environment_vars
from zenml.stack import Flavor, Stack, StackComponent, StackComponentConfig
from zenml.steps.step_context import RunContext, get_or_create_run_context
from zenml.utils.env_utils import temporary_environment
from zenml.utils.pydantic_utils import before_validator_handler

if TYPE_CHECKING:
    from zenml.config.step_configurations import Step
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotResponse,
        ScheduleResponse,
        ScheduleUpdate,
        StepRunResponse,
    )

logger = get_logger(__name__)


class SubmissionResult:
    """Result of submitting a pipeline run."""

    def __init__(
        self,
        wait_for_completion: Optional[Callable[[], None]] = None,
        metadata: Optional[Dict[str, MetadataType]] = None,
    ):
        """Initialize a submission result.

        Args:
            wait_for_completion: A function that waits for the pipeline run to
                complete. If provided, this will be called after the pipeline
                run was submitted successfully.
            metadata: Metadata for the pipeline run or schedule.
        """
        self.wait_for_completion = wait_for_completion
        self.metadata = metadata


class BaseOrchestratorConfig(StackComponentConfig):
    """Base orchestrator config."""

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _deprecations(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        """Validate and/or remove deprecated fields.

        Args:
            data: The values to validate.

        Returns:
            The validated values.
        """
        if "custom_docker_base_image_name" in data:
            image_name = data.pop("custom_docker_base_image_name", None)
            if image_name:
                logger.warning(
                    "The 'custom_docker_base_image_name' field has been "
                    "deprecated. To use a custom base container image with your "
                    "orchestrators, please use the DockerSettings in your "
                    "pipeline (see https://docs.zenml.io/concepts/containerization)."
                )

        return data

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return False

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return False

    @property
    def supports_client_side_caching(self) -> bool:
        """Whether the orchestrator supports client side caching.

        Returns:
            Whether the orchestrator supports client side caching.
        """
        return True

    @property
    def handles_step_retries(self) -> bool:
        """Whether the orchestrator handles step retries.

        Returns:
            Whether the orchestrator handles step retries.
        """
        return False


class BaseOrchestrator(StackComponent, ABC):
    """Base class for all orchestrators."""

    _active_snapshot: Optional["PipelineSnapshotResponse"] = None

    @property
    def config(self) -> BaseOrchestratorConfig:
        """Returns the `BaseOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(BaseOrchestratorConfig, self._config)

    @abstractmethod
    def get_orchestrator_run_id(self) -> str:
        """Returns the run id of the active orchestrator run.

        Important:
        - This needs to be a unique ID and return the same value for
        all steps of a pipeline run.
        - For dynamic pipelines, this needs to be a unique ID which will be
        fetched in the orchestration environment. If the orchestrator supports
        retries of the orchestration environment, this ID needs to be the same
        for all retries.

        Returns:
            The orchestrator run id.
        """

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Returns:
            Optional submission result.
        """
        return None

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment.
            placeholder_run: An optional placeholder run.

        Returns:
            Optional submission result.
        """
        return None

    def prepare_or_run_pipeline(
        self,
        deployment: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[Iterator[Dict[str, MetadataType]]]:
        """DEPRECATED: Prepare or run a pipeline.

        Args:
            deployment: The deployment to prepare or run.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment. These don't need to be set if running locally.
            placeholder_run: An optional placeholder run for the deployment.
        """

    def run(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> None:
        """Runs a pipeline on a stack.

        Args:
            snapshot: The pipeline snapshot.
            stack: The stack on which to run the pipeline.
            placeholder_run: An optional placeholder run for the snapshot.
                This will be deleted in case the pipeline run failed.

        Raises:
            RunMonitoringError: If a failure happened while monitoring the
                pipeline run.
            PipelineSubmissionError: If a pipeline submission failed on the
                orchestrator.
        """
        self._prepare_run(snapshot=snapshot)

        pipeline_run_id: Optional[UUID] = None
        schedule_id: Optional[UUID] = None
        if snapshot.schedule:
            schedule_id = snapshot.schedule.id
        if placeholder_run:
            pipeline_run_id = placeholder_run.id

        base_environment, secrets = get_config_environment_vars(
            schedule_id=schedule_id,
            pipeline_run_id=pipeline_run_id,
        )

        # TODO: for now, we don't support separate secrets from environment
        # in the orchestrator environment
        base_environment.update(secrets)

        if placeholder_run and placeholder_run.original_run:
            from zenml.execution.pipeline.utils import (
                skip_steps_and_prune_snapshot,
            )

            run_required = skip_steps_and_prune_snapshot(
                snapshot=snapshot,
                pipeline_run=placeholder_run,
            )

            if not run_required:
                self._cleanup_run()
                return

        prevent_client_side_caching = handle_bool_env_var(
            ENV_ZENML_PREVENT_CLIENT_SIDE_CACHING, default=False
        )

        if (
            placeholder_run
            and self.config.supports_client_side_caching
            and not snapshot.schedule
            and not snapshot.is_dynamic
            and not prevent_client_side_caching
        ):
            from zenml.orchestrators import cache_utils

            run_required = (
                cache_utils.create_cached_step_runs_and_prune_snapshot(
                    snapshot=snapshot,
                    pipeline_run=placeholder_run,
                    stack=stack,
                )
            )

            if not run_required:
                self._cleanup_run()
                return
        else:
            logger.debug("Skipping client-side caching.")

        try:
            if (
                not snapshot.is_dynamic
                and getattr(self.submit_pipeline, "__func__", None)
                is BaseOrchestrator.submit_pipeline
            ):
                logger.warning(
                    "The orchestrator '%s' is still using the deprecated "
                    "`prepare_or_run_pipeline(...)` method which will be "
                    "removed in the future. Please implement the replacement "
                    "`submit_pipeline(...)` method for your custom "
                    "orchestrator.",
                    self.name,
                )
                if metadata_iterator := self.prepare_or_run_pipeline(
                    deployment=snapshot,
                    stack=stack,
                    environment=base_environment,
                    placeholder_run=placeholder_run,
                ):
                    for metadata_dict in metadata_iterator:
                        try:
                            if placeholder_run:
                                publish_pipeline_run_metadata(
                                    pipeline_run_id=placeholder_run.id,
                                    pipeline_run_metadata={
                                        self.id: metadata_dict
                                    },
                                )
                        except Exception as e:
                            logger.debug(
                                "Something went went wrong trying to publish the"
                                f"run metadata: {e}"
                            )
            else:
                if snapshot.is_dynamic:
                    submission_result = self.submit_dynamic_pipeline(
                        snapshot=snapshot,
                        stack=stack,
                        environment=base_environment,
                        placeholder_run=placeholder_run,
                    )
                else:
                    step_environments = {}
                    for (
                        invocation_id,
                        step,
                    ) in snapshot.step_configurations.items():
                        from zenml.utils.env_utils import (
                            get_runtime_environment,
                        )

                        step_environment = get_runtime_environment(
                            config=step.config,
                            stack=stack,
                        )

                        combined_environment = base_environment.copy()
                        combined_environment.update(step_environment)
                        step_environments[invocation_id] = combined_environment

                    submission_result = self.submit_pipeline(
                        snapshot=snapshot,
                        stack=stack,
                        base_environment=base_environment,
                        step_environments=step_environments,
                        placeholder_run=placeholder_run,
                    )

                if placeholder_run:
                    publish_pipeline_run_status_update(
                        pipeline_run_id=placeholder_run.id,
                        status=ExecutionStatus.PROVISIONING,
                    )

                if submission_result:
                    if submission_result.metadata:
                        if placeholder_run:
                            try:
                                publish_pipeline_run_metadata(
                                    pipeline_run_id=placeholder_run.id,
                                    pipeline_run_metadata={
                                        self.id: submission_result.metadata
                                    },
                                )
                            except Exception as e:
                                logger.debug(
                                    "Error publishing run metadata: %s", e
                                )
                        elif snapshot.schedule:
                            try:
                                publish_schedule_metadata(
                                    schedule_id=snapshot.schedule.id,
                                    schedule_metadata={
                                        self.id: submission_result.metadata
                                    },
                                )
                            except Exception as e:
                                logger.debug(
                                    "Error publishing schedule metadata: %s", e
                                )

                    if submission_result.wait_for_completion:
                        try:
                            submission_result.wait_for_completion()
                        except KeyboardInterrupt as e:
                            message = (
                                "Run monitoring interrupted, but "
                                "the pipeline is still executing."
                            )
                            if placeholder_run:
                                message += (
                                    " If you want to stop the run, use: `zenml "
                                    f"pipeline runs stop {placeholder_run.id}`"
                                )
                            # TODO: once we don't support Python 3.10 anymore,
                            # use `exception.add_note` instead.
                            e.args = (message,)
                            raise RunMonitoringError(original_exception=e)
                        except BaseException as e:
                            raise RunMonitoringError(original_exception=e)
        except PipelineSubmissionError as e:
            # clean-up actions in case of failure

            if snapshot.schedule:
                # delete created DB schedules
                Client().zen_store.delete_schedule(
                    snapshot.schedule.id, soft=False
                )

            raise e
        finally:
            self._cleanup_run()

    def resume_run(
        self,
        snapshot: "PipelineSnapshotResponse",
        run: "PipelineRunResponse",
        stack: "Stack",
        force_async: bool = False,
    ) -> None:
        """Resume an existing dynamic pipeline run.

        Args:
            snapshot: Snapshot backing the run.
            run: Existing pipeline run to resume.
            stack: Stack used for the resume.
            force_async: Whether to skip waiting for submission completion.

        Raises:
            RuntimeError: If the snapshot does not represent a dynamic pipeline.
        """
        if not snapshot.is_dynamic:
            raise RuntimeError("Cannot resume a non-dynamic pipeline.")

        self._prepare_run(snapshot=snapshot)

        base_environment, secrets = get_config_environment_vars(
            pipeline_run_id=run.id,
        )
        base_environment.update(secrets)

        try:
            submission_result = self.submit_dynamic_pipeline(
                snapshot=snapshot,
                stack=stack,
                environment=base_environment,
                placeholder_run=run,
            )

            if submission_result:
                if submission_result.metadata:
                    # TODO: Do we somehow prefix the metadata here
                    pass

                if submission_result.wait_for_completion and not force_async:
                    submission_result.wait_for_completion()
        finally:
            self._cleanup_run()

    def run_step(
        self,
        step: "Step",
    ) -> None:
        """Runs the given step.

        Args:
            step: The step to run.
        """
        from zenml.execution.step.utils import launch_step

        assert self._active_snapshot

        launch_step(
            snapshot=self._active_snapshot,
            step=step,
            orchestrator_run_id=self.get_orchestrator_run_id(),
            retry=not self.config.handles_step_retries,
        )

    @property
    def supports_dynamic_pipelines(self) -> bool:
        """Whether the orchestrator supports dynamic pipelines.

        Returns:
            Whether the orchestrator supports dynamic pipelines.
        """
        return (
            getattr(self.submit_dynamic_pipeline, "__func__", None)
            is not BaseOrchestrator.submit_dynamic_pipeline
        )

    @property
    def can_run_isolated_steps(self) -> bool:
        """Whether the orchestrator can run isolated steps.

        Returns:
            Whether the orchestrator can run isolated steps.
        """
        return (
            getattr(self.submit_isolated_step, "__func__", None)
            is not BaseOrchestrator.submit_isolated_step
        )

    def submit_isolated_step(
        self, step_run_info: "StepRunInfo", environment: Dict[str, str]
    ) -> None:
        """Submit an isolated step.

        Args:
            step_run_info: The step run information.
            environment: The environment variables to set in the execution
                environment.

        Raises:
            NotImplementedError: If the orchestrator does not implement this
                method.
        """
        raise NotImplementedError(
            "Submitting isolated steps is not implemented for "
            f"the {self.__class__.__name__} orchestrator."
        )

    def get_isolated_step_status(
        self, step_run: "StepRunResponse"
    ) -> ExecutionStatus:
        """Get the status of an isolated step run.

        Args:
            step_run: The step run.

        Raises:
            NotImplementedError: If the orchestrator does not implement this
                method.
        """
        raise NotImplementedError(
            "Getting the status of isolated steps is not implemented for "
            f"the {self.__class__.__name__} orchestrator."
        )

    def wait_for_isolated_step(
        self, step_run: "StepRunResponse"
    ) -> ExecutionStatus:
        """Wait for an isolated step run to complete.

        Args:
            step_run: The step run.

        Returns:
            The final status of the isolated step run.
        """
        sleep_interval = 1
        max_sleep_interval = 16

        while True:
            try:
                status = self.get_isolated_step_status(step_run)
            except Exception as e:
                logger.error(
                    "Failed to get orchestrator status of isolated step "
                    "`%s`: %s",
                    step_run.id,
                    e,
                )
                # Fall back to the status of the ZenML server
                status = (
                    Client().get_run_step(step_run.id, hydrate=False).status
                )

            if status.is_finished or status == ExecutionStatus.RETRYING:
                return status

            logger.debug(
                "Waiting for isolated step with ID %s to finish (current "
                "status: %s)",
                step_run.id,
                status,
            )
            time.sleep(sleep_interval)
            if sleep_interval < max_sleep_interval:
                sleep_interval *= 2

    @property
    def can_stop_isolated_steps(self) -> bool:
        """Whether the orchestrator can stop isolated steps.

        Returns:
            Whether the orchestrator can stop isolated steps.
        """
        return (
            getattr(self.stop_isolated_step, "__func__", None)
            is not BaseOrchestrator.stop_isolated_step
        )

    def stop_isolated_step(self, step_run: "StepRunResponse") -> None:
        """Stop an isolated step run.

        Args:
            step_run: The step run to stop.

        Raises:
            NotImplementedError: If the orchestrator does not implement this
                method.
        """
        raise NotImplementedError(
            "Stopping isolated steps is not implemented for "
            f"the {self.__class__.__name__} orchestrator."
        )

    @staticmethod
    def requires_resources_in_orchestration_environment(
        step: "Step",
    ) -> bool:
        """Checks if the orchestrator should run this step on special resources.

        Args:
            step: The step that will be checked.

        Returns:
            True if the step requires special resources in the orchestration
            environment, False otherwise.
        """
        # If the step requires custom resources and doesn't run with a step
        # operator, it would need these requirements in the orchestrator
        # environment
        if step.config.step_operator:
            return False

        return not step.config.resource_settings.empty

    def _prepare_run(self, snapshot: "PipelineSnapshotResponse") -> None:
        """Prepares a run.

        Args:
            snapshot: The snapshot to prepare.
        """
        self._validate_execution_mode(snapshot)
        self._active_snapshot = snapshot

    def _cleanup_run(self) -> None:
        """Cleans up the active run."""
        self._active_snapshot = None

    @property
    def supported_execution_modes(self) -> List[ExecutionMode]:
        """Returns the supported execution modes for this flavor.

        Returns:
            A tuple of supported execution modes.
        """
        return [ExecutionMode.CONTINUE_ON_FAILURE]

    @property
    def run_init_cleanup_at_step_level(self) -> bool:
        """Whether the orchestrator runs the init and cleanup hooks at step level.

        For orchestrators that run their steps in isolated step environments,
        the run context cannot be shared between steps. In this case, the init
        and cleanup hooks need to be run at step level for each individual step.

        For orchestrators that run their steps in a shared environment with a
        shared memory (e.g. the local orchestrator), the init and cleanup hooks
        can be run at run level and this property should be overridden to return
        True.

        Returns:
            Whether the orchestrator runs the init and cleanup hooks at step
            level.
        """
        return True

    @classmethod
    def run_init_hook(cls, snapshot: "PipelineSnapshotResponse") -> None:
        """Runs the init hook.

        Args:
            snapshot: The snapshot to run the init hook for.

        Raises:
            HookExecutionException: If the init hook fails.
        """
        # The lifetime of the run context starts when the init hook is executed
        # and ends when the cleanup hook is executed
        run_context = get_or_create_run_context()
        init_hook_source = snapshot.pipeline_configuration.init_hook_source
        init_hook_kwargs = snapshot.pipeline_configuration.init_hook_kwargs

        # We only run the init hook once, if the (thread-local) run context
        # associated with the current run has not been initialized yet. This
        # allows us to run the init hook only once per run per execution
        # environment (process, container, etc.).
        if not run_context.initialized:
            if not init_hook_source:
                run_context.initialize(None)
                return

            logger.info("Executing the pipeline's init hook...")
            try:
                with temporary_environment(
                    snapshot.pipeline_configuration.environment
                ):
                    run_state = load_and_run_hook(
                        init_hook_source,
                        hook_parameters=init_hook_kwargs,
                        raise_on_error=True,
                    )
            except Exception as e:
                raise HookExecutionException(
                    f"Failed to execute init hook for pipeline "
                    f"{snapshot.pipeline_configuration.name}"
                ) from e

            run_context.initialize(run_state)

    @classmethod
    def run_cleanup_hook(cls, snapshot: "PipelineSnapshotResponse") -> None:
        """Runs the cleanup hook.

        Args:
            snapshot: The snapshot to run the cleanup hook for.
        """
        # The lifetime of the run context starts when the init hook is executed
        # and ends when the cleanup hook is executed
        if not RunContext._exists():
            return

        if (
            cleanup_hook_source
            := snapshot.pipeline_configuration.cleanup_hook_source
        ):
            logger.info("Executing the pipeline's cleanup hook...")
            with temporary_environment(
                snapshot.pipeline_configuration.environment
            ):
                load_and_run_hook(
                    cleanup_hook_source,
                    raise_on_error=False,
                )

        # Destroy the run context, so it's created anew for the next run
        RunContext._clear()

    def _validate_execution_mode(
        self, snapshot: "PipelineSnapshotResponse"
    ) -> None:
        """Validate that the requested execution mode is supported.

        Args:
            snapshot: The snapshot to validate.

        Raises:
            ValueError: If the execution mode is not supported.
        """
        if snapshot.is_dynamic:
            # We can't validate execution modes for dynamic pipelines yet
            return

        execution_mode = snapshot.pipeline_configuration.execution_mode

        if execution_mode not in self.supported_execution_modes:
            raise ValueError(
                f"Execution mode {execution_mode} is not supported by the "
                f"{self.__class__.__name__} orchestrator."
            )

    def fetch_status(
        self, run: "PipelineRunResponse", include_steps: bool = False
    ) -> Tuple[
        Optional[ExecutionStatus], Optional[Dict[str, ExecutionStatus]]
    ]:
        """Refreshes the status of a specific pipeline run.

        Args:
            run: A pipeline run response to fetch its status.
            include_steps: If True, also fetch the status of individual steps.

        Raises:
            NotImplementedError: If any orchestrator inheriting from the base
                class does not implement this logic.
        """
        raise NotImplementedError(
            "The fetch status functionality is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    def stop_run(
        self, run: "PipelineRunResponse", graceful: bool = False
    ) -> None:
        """Stops a specific pipeline run.

        This method should only be called if the orchestrator's
        supports_cancellation property is True.

        Args:
            run: A pipeline run response to stop.
            graceful: If True, allows for graceful shutdown where possible.
                If False, forces immediate termination. Default is False.

        Raises:
            IllegalOperationError: If the run has no orchestrator run id yet.
        """
        if not run.orchestrator_run_id:
            raise IllegalOperationError(
                "Cannot stop a pipeline run that has no orchestrator run id "
                "yet."
            )

        # Update pipeline status to STOPPING before calling concrete implementation
        # Initiates graceful termination.
        publish_pipeline_run_status_update(
            pipeline_run_id=run.id,
            status=ExecutionStatus.STOPPING,
            status_reason="Manual stop requested.",
        )

        # Now call the concrete implementation
        self._stop_run(run=run, graceful=graceful)

    def _stop_run(
        self, run: "PipelineRunResponse", graceful: bool = False
    ) -> None:
        """Concrete implementation of pipeline stopping logic.

        This method should be implemented by concrete orchestrator classes
        instead of stop_run to ensure proper status management.

        Args:
            run: A pipeline run response to stop (already updated to STOPPING status).
            graceful: If True, allows for graceful shutdown where possible.
                If False, forces immediate termination. Default is True.
        """
        if graceful:
            # This should work out of the box for HeartBeat step termination.
            # Orchestrators should extend the functionality to cover other scenarios.
            self._stop_run_gracefully(pipeline_run=run)
        else:
            self._stop_run_forcefully(pipeline_run=run)

    def _stop_run_gracefully(
        self, pipeline_run: "PipelineRunResponse"
    ) -> None:
        pass

    def _stop_run_forcefully(
        self, pipeline_run: "PipelineRunResponse"
    ) -> None:
        raise NotImplementedError(
            "The forceful stop run functionality is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    @property
    def supports_schedule_updates(self) -> bool:
        """Whether the orchestrator supports updating schedules.

        Returns:
            Whether the orchestrator supports updating schedules.
        """
        return (
            getattr(self.update_schedule, "__func__", None)
            is not BaseOrchestrator.update_schedule
        )

    @property
    def supports_schedule_deletion(self) -> bool:
        """Whether the orchestrator supports deleting schedules.

        Returns:
            Whether the orchestrator supports deleting schedules.
        """
        return (
            getattr(self.delete_schedule, "__func__", None)
            is not BaseOrchestrator.delete_schedule
        )

    def update_schedule(
        self, schedule: "ScheduleResponse", update: "ScheduleUpdate"
    ) -> None:
        """Updates a schedule.

        Args:
            schedule: The schedule to update.
            update: The update to apply to the schedule.

        Raises:
            NotImplementedError: If the functionality is not implemented.
        """
        raise NotImplementedError(
            "Schedule updating is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    def delete_schedule(self, schedule: "ScheduleResponse") -> None:
        """Deletes a schedule.

        Args:
            schedule: The schedule to delete.

        Raises:
            NotImplementedError: If the functionality is not implemented.
        """
        raise NotImplementedError(
            "Schedule deletion is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )


class BaseOrchestratorFlavor(Flavor):
    """Base orchestrator flavor class."""

    @property
    def type(self) -> StackComponentType:
        """Returns the flavor type.

        Returns:
            The flavor type.
        """
        return StackComponentType.ORCHESTRATOR

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return BaseOrchestratorConfig

    @property
    @abstractmethod
    def implementation_class(self) -> Type["BaseOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of pipeline run tables."""

import json
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence
from uuid import UUID

from pydantic import ConfigDict
from sqlalchemy import String, UniqueConstraint
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlalchemy.orm import Session, joinedload, object_session, selectinload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import TEXT, Column, Field, Relationship, col, select

from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.pipeline_spec import PipelineSpec
from zenml.config.step_configurations import Step
from zenml.constants import MEDIUMTEXT_MAX_LENGTH, TEXT_FIELD_MAX_LENGTH
from zenml.enums import (
    ExecutionMode,
    ExecutionStatus,
    MetadataResourceTypes,
    PipelineRunTriggeredByType,
    RunWaitConditionStatus,
    TaggableResourceTypes,
    VisualizationResourceTypes,
)
from zenml.exceptions import IllegalOperationError
from zenml.logger import get_logger
from zenml.models import (
    PipelineResponse,
    PipelineRunRequest,
    PipelineRunResponse,
    PipelineRunResponseBody,
    PipelineRunResponseMetadata,
    PipelineRunTriggerInfo,
    PipelineRunUpdate,
    RunMetadataEntry,
)
from zenml.models.v2.core.pipeline_run import PipelineRunResponseResources
from zenml.utils.run_utils import (
    build_dag,
    find_all_downstream_steps,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.constants import MODEL_VERSION_TABLENAME
from zenml.zen_stores.schemas.pipeline_build_schemas import PipelineBuildSchema
from zenml.zen_stores.schemas.pipeline_schemas import PipelineSchema
from zenml.zen_stores.schemas.pipeline_snapshot_schemas import (
    PipelineSnapshotSchema,
)
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schedule_schema import ScheduleSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
)
from zenml.zen_stores.schemas.stack_schemas import StackSchema
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import (
    RunMetadataInterface,
    jl_arg,
)

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.curated_visualization_schemas import (
        CuratedVisualizationSchema,
    )
    from zenml.zen_stores.schemas.logs_schemas import LogsSchema
    from zenml.zen_stores.schemas.model_schemas import (
        ModelVersionPipelineRunSchema,
        ModelVersionSchema,
    )
    from zenml.zen_stores.schemas.run_metadata_schemas import RunMetadataSchema
    from zenml.zen_stores.schemas.run_wait_condition_schemas import (
        RunWaitConditionSchema,
    )
    from zenml.zen_stores.schemas.service_schemas import ServiceSchema
    from zenml.zen_stores.schemas.step_run_schemas import StepRunSchema
    from zenml.zen_stores.schemas.tag_schemas import TagSchema
    from zenml.zen_stores.schemas.trigger_schemas import TriggerSchema


logger = get_logger(__name__)


class PipelineRunSchema(NamedSchema, RunMetadataInterface, table=True):
    """SQL Model for pipeline runs."""

    __tablename__ = "pipeline_run"
    __table_args__ = (
        UniqueConstraint(
            "snapshot_id",
            "orchestrator_run_id",
            name="unique_orchestrator_run_id_for_snapshot_id",
        ),
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_run_name_in_project",
        ),
    )

    # Fields
    orchestrator_run_id: Optional[str] = Field(nullable=True)
    start_time: Optional[datetime] = Field(nullable=True)
    end_time: Optional[datetime] = Field(nullable=True, default=None)
    in_progress: bool = Field(nullable=False)
    status: str = Field(nullable=False)
    status_reason: Optional[str] = Field(nullable=True)
    orchestrator_environment: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )
    index: int = Field(nullable=False)
    enable_heartbeat: bool = Field(nullable=False)
    exception_info: Optional[str] = Field(
        sa_column=Column(
            String(length=MEDIUMTEXT_MAX_LENGTH).with_variant(
                MEDIUMTEXT, "mysql"
            ),
            nullable=True,
        )
    )

    # Foreign keys
    snapshot_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineSnapshotSchema.__tablename__,
        source_column="snapshot_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=True,
    )
    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    pipeline_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineSchema.__tablename__,
        source_column="pipeline_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    model_version_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=MODEL_VERSION_TABLENAME,
        source_column="model_version_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    # Relationships
    snapshot: Optional["PipelineSnapshotSchema"] = Relationship(
        back_populates="pipeline_runs"
    )
    project: "ProjectSchema" = Relationship(back_populates="runs")
    user: Optional["UserSchema"] = Relationship(back_populates="runs")
    run_metadata: List["RunMetadataSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            secondary="run_metadata_resource",
            primaryjoin=f"and_(foreign(RunMetadataResourceSchema.resource_type)=='{MetadataResourceTypes.PIPELINE_RUN.value}', foreign(RunMetadataResourceSchema.resource_id)==PipelineRunSchema.id)",
            secondaryjoin="RunMetadataSchema.id==foreign(RunMetadataResourceSchema.run_metadata_id)",
            overlaps="run_metadata",
        ),
    )
    logs: List["LogsSchema"] = Relationship(
        back_populates="pipeline_run",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    step_runs: List["StepRunSchema"] = Relationship(
        sa_relationship_kwargs={
            "cascade": "delete",
            "order_by": "asc(StepRunSchema.start_time)",
        },
    )
    wait_conditions: List["RunWaitConditionSchema"] = Relationship(
        back_populates="run",
        sa_relationship_kwargs={
            "cascade": "delete",
            "order_by": "desc(RunWaitConditionSchema.created)",
        },
    )
    model_version: "ModelVersionSchema" = Relationship(
        back_populates="pipeline_runs",
    )

    # Temporary fields and foreign keys to be deprecated
    pipeline_configuration: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )
    client_environment: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )

    stack_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=StackSchema.__tablename__,
        source_column="stack_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    build_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineBuildSchema.__tablename__,
        source_column="build_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    schedule_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=ScheduleSchema.__tablename__,
        source_column="schedule_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    original_run_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=__tablename__,
        source_column="original_run_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    original_run: Optional["PipelineRunSchema"] = Relationship(
        sa_relationship_kwargs={"remote_side": "PipelineRunSchema.id"}
    )

    stack: Optional["StackSchema"] = Relationship()
    build: Optional["PipelineBuildSchema"] = Relationship()
    schedule: Optional["ScheduleSchema"] = Relationship()
    pipeline: Optional["PipelineSchema"] = Relationship()
    triggered_by: Optional[UUID] = None
    triggered_by_type: Optional[str] = None

    services: List["ServiceSchema"] = Relationship(
        back_populates="pipeline_run",
    )
    tags: List["TagSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=f"and_(foreign(TagResourceSchema.resource_type)=='{TaggableResourceTypes.PIPELINE_RUN.value}', foreign(TagResourceSchema.resource_id)==PipelineRunSchema.id)",
            secondary="tag_resource",
            secondaryjoin="TagSchema.id == foreign(TagResourceSchema.tag_id)",
            order_by="TagSchema.name",
            overlaps="tags",
        ),
    )
    visualizations: List["CuratedVisualizationSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=(
                "and_(CuratedVisualizationSchema.resource_type"
                f"=='{VisualizationResourceTypes.PIPELINE_RUN.value}', "
                "foreign(CuratedVisualizationSchema.resource_id)==PipelineRunSchema.id)"
            ),
            overlaps="visualizations",
            cascade="delete",
            order_by="CuratedVisualizationSchema.display_order",
        ),
    )

    # PipelineRun -> Trigger directly (through trigger_execution)
    trigger: Optional["TriggerSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin="PipelineRunSchema.id == TriggerExecutionSchema.pipeline_run_id",
            secondary="trigger_execution",
            secondaryjoin="TriggerExecutionSchema.trigger_id == TriggerSchema.id",
            uselist=False,  # We expect 1 or 0 triggers exactly.
            viewonly=True,  # Avoid write ambiguity, write via TriggerExecutionSchema
        )
    )

    # Needed for cascade deletion
    model_versions_pipeline_runs_links: List[
        "ModelVersionPipelineRunSchema"
    ] = Relationship(
        back_populates="pipeline_run",
        sa_relationship_kwargs={"cascade": "delete"},
    )

    model_config = ConfigDict(protected_namespaces=())  # type: ignore[assignment]

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        from zenml.zen_stores.schemas import ModelVersionSchema

        options = []

        # if include_metadata:
        #     options.extend(
        #         [
        #             joinedload(jl_arg(PipelineRunSchema.run_metadata)),
        #         ]
        #     )

        if include_resources:
            options.extend(
                [
                    selectinload(
                        jl_arg(PipelineRunSchema.model_version)
                    ).joinedload(
                        jl_arg(ModelVersionSchema.model), innerjoin=True
                    ),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(
                        jl_arg(PipelineSnapshotSchema.source_snapshot)
                    ),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(jl_arg(PipelineSnapshotSchema.pipeline)),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(jl_arg(PipelineSnapshotSchema.stack)),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(jl_arg(PipelineSnapshotSchema.build)),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(jl_arg(PipelineSnapshotSchema.schedule)),
                    selectinload(
                        jl_arg(PipelineRunSchema.snapshot)
                    ).joinedload(
                        jl_arg(PipelineSnapshotSchema.code_reference)
                    ),
                    selectinload(jl_arg(PipelineRunSchema.logs)),
                    selectinload(jl_arg(PipelineRunSchema.wait_conditions)),
                    selectinload(jl_arg(PipelineRunSchema.user)),
                    selectinload(jl_arg(PipelineRunSchema.tags)),
                    selectinload(jl_arg(PipelineRunSchema.visualizations)),
                    joinedload(jl_arg(PipelineRunSchema.trigger)),
                ]
            )

        return options

    @classmethod
    def from_request(
        cls,
        request: "PipelineRunRequest",
        pipeline_id: UUID,
        index: int,
        enable_heartbeat: bool,
    ) -> "PipelineRunSchema":
        """Convert a `PipelineRunRequest` to a `PipelineRunSchema`.

        Args:
            request: The request to convert.
            pipeline_id: The ID of the pipeline.
            index: The index of the pipeline run.
            enable_heartbeat: Whether the heartbeat should be enabled.

        Returns:
            The created `PipelineRunSchema`.
        """
        orchestrator_environment = json.dumps(request.orchestrator_environment)
        if len(orchestrator_environment) > TEXT_FIELD_MAX_LENGTH:
            logger.warning(
                "Orchestrator environment is too large to be stored in the "
                "database. Skipping."
            )
            orchestrator_environment = "{}"

        triggered_by = None
        triggered_by_type = None
        if request.trigger_info:
            if request.trigger_info.step_run_id:
                triggered_by = request.trigger_info.step_run_id
                triggered_by_type = PipelineRunTriggeredByType.STEP_RUN.value
            elif request.trigger_info.deployment_id:
                triggered_by = request.trigger_info.deployment_id
                triggered_by_type = PipelineRunTriggeredByType.DEPLOYMENT.value

        return cls(
            project_id=request.project,
            user_id=request.user,
            name=request.name,
            orchestrator_run_id=request.orchestrator_run_id,
            orchestrator_environment=orchestrator_environment,
            start_time=request.start_time,
            end_time=request.end_time,
            status=request.status.value,
            index=index,
            in_progress=not request.status.is_finished,
            status_reason=request.status_reason,
            pipeline_id=pipeline_id,
            snapshot_id=request.snapshot,
            triggered_by=triggered_by,
            triggered_by_type=triggered_by_type,
            enable_heartbeat=enable_heartbeat,
            exception_info=request.exception_info.model_dump_json()
            if request.exception_info
            else None,
            original_run_id=request.original_run_id,
        )

    def get_pipeline_configuration(self) -> PipelineConfiguration:
        """Get the pipeline configuration for the pipeline run.

        Raises:
            RuntimeError: if the pipeline run has no snapshot and no pipeline
                configuration.

        Returns:
            The pipeline configuration.
        """
        if self.snapshot:
            pipeline_config = PipelineConfiguration.model_validate_json(
                self.snapshot.pipeline_configuration
            )
        elif self.pipeline_configuration:
            pipeline_config = PipelineConfiguration.model_validate_json(
                self.pipeline_configuration
            )
        else:
            raise RuntimeError(
                "Pipeline run has no snapshot and no pipeline configuration."
            )

        pipeline_config.finalize_substitutions(
            start_time=self.start_time, inplace=True
        )
        return pipeline_config

    def get_step_configuration(self, step_name: str) -> Step:
        """Get the step configuration for the pipeline run.

        Args:
            step_name: The name of the step to get the configuration for.

        Raises:
            RuntimeError: If the pipeline run has no snapshot.

        Returns:
            The step configuration.
        """
        if self.snapshot:
            pipeline_configuration = self.get_pipeline_configuration()
            return Step.from_dict(
                data=json.loads(
                    self.snapshot.get_step_configuration(step_name).config
                ),
                pipeline_configuration=pipeline_configuration,
            )
        else:
            raise RuntimeError("Pipeline run has no snapshot.")

    def get_upstream_steps(self) -> Dict[str, List[str]]:
        """Get the list of all the upstream steps for each step.

        Returns:
            The list of upstream steps for each step.

        Raises:
            RuntimeError: If the pipeline run has no snapshot or
                the snapshot has no pipeline spec.
        """
        if self.snapshot and self.snapshot.pipeline_spec:
            pipeline_spec = PipelineSpec.model_validate_json(
                self.snapshot.pipeline_spec
            )
            steps = {}
            for step_spec in pipeline_spec.steps:
                steps[step_spec.invocation_id] = step_spec.upstream_steps
            return steps
        else:
            raise RuntimeError("Pipeline run has no snapshot.")

    def _get_step_run_statuses(self) -> List[ExecutionStatus]:
        """Get the statuses of the steps in the pipeline run.

        Raises:
            RuntimeError: If the pipeline run has no session.

        Returns:
            The statuses of the steps in the pipeline run.
        """
        from zenml.zen_stores.schemas.step_run_schemas import StepRunSchema

        if session := object_session(self):
            step_run_statuses = session.execute(
                select(StepRunSchema.status)
                .where(StepRunSchema.pipeline_run_id == self.id)
                .where(
                    col(StepRunSchema.status) != ExecutionStatus.RETRIED.value
                )
            ).scalars()

            return [ExecutionStatus(status) for status in step_run_statuses]
        else:
            raise RuntimeError(
                "Missing DB session to fetch step run statuses."
            )

    def fetch_metadata_collection(
        self, include_full_metadata: bool = False, **kwargs: Any
    ) -> Dict[str, List[RunMetadataEntry]]:
        """Fetches all the metadata entries related to the pipeline run.

        Args:
            include_full_metadata: Whether the full metadata will be included.
            **kwargs: Keyword arguments.

        Returns:
            a dictionary, where the key is the key of the metadata entry
                and the values represent the list of entries with this key.
        """
        # Fetch the metadata related to this run
        metadata_collection = super().fetch_metadata_collection(**kwargs)

        if include_full_metadata:
            # Fetch the metadata related to the steps of this run
            for s in self.step_runs:
                step_metadata = s.fetch_metadata_collection()
                for k, v in step_metadata.items():
                    metadata_collection[f"{s.name}::{k}"] = v

            # Fetch the metadata related to the schedule of this run
            if self.snapshot is not None:
                if schedule := self.snapshot.schedule:
                    schedule_metadata = schedule.fetch_metadata_collection()
                    for k, v in schedule_metadata.items():
                        metadata_collection[f"schedule:{k}"] = v

        return metadata_collection

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        include_python_packages: bool = False,
        include_full_metadata: bool = False,
        **kwargs: Any,
    ) -> "PipelineRunResponse":
        """Convert a `PipelineRunSchema` to a `PipelineRunResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            include_python_packages: Whether the python packages will be filled.
            include_full_metadata: Whether the full metadata will be included.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The created `PipelineRunResponse`.

        Raises:
            RuntimeError: if the model creation fails.
        """
        if self.snapshot is not None:
            config = PipelineConfiguration.model_validate_json(
                self.snapshot.pipeline_configuration
            )
            client_environment = json.loads(self.snapshot.client_environment)
        elif self.pipeline_configuration is not None:
            config = PipelineConfiguration.model_validate_json(
                self.pipeline_configuration
            )
            client_environment = (
                json.loads(self.client_environment)
                if self.client_environment
                else {}
            )
        else:
            raise RuntimeError(
                "Pipeline run model creation has failed. Each pipeline run "
                "entry should either have a snapshot_id or "
                "pipeline_configuration."
            )

        config.finalize_substitutions(start_time=self.start_time, inplace=True)

        body = PipelineRunResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            status=ExecutionStatus(self.status),
            status_reason=self.status_reason,
            created=self.created,
            updated=self.updated,
            in_progress=self.in_progress,
            index=self.index,
        )
        metadata = None
        if include_metadata:
            is_templatable = False
            if (
                self.snapshot
                and self.snapshot.build
                and not self.snapshot.build.is_local
                and self.snapshot.build.stack_id
            ):
                is_templatable = True

            orchestrator_environment = (
                json.loads(self.orchestrator_environment)
                if self.orchestrator_environment
                else {}
            )

            if not include_python_packages:
                client_environment.pop("python_packages", None)
                orchestrator_environment.pop("python_packages", None)

            trigger_info: Optional[PipelineRunTriggerInfo] = None
            if self.triggered_by and self.triggered_by_type:
                if (
                    self.triggered_by_type
                    == PipelineRunTriggeredByType.STEP_RUN.value
                ):
                    trigger_info = PipelineRunTriggerInfo(
                        step_run_id=self.triggered_by,
                    )
                elif (
                    self.triggered_by_type
                    == PipelineRunTriggeredByType.DEPLOYMENT.value
                ):
                    trigger_info = PipelineRunTriggerInfo(
                        deployment_id=self.triggered_by,
                    )

            metadata = PipelineRunResponseMetadata(
                run_metadata=self.fetch_metadata(
                    include_full_metadata=include_full_metadata
                ),
                config=config,
                start_time=self.start_time,
                end_time=self.end_time,
                client_environment=client_environment,
                orchestrator_environment=orchestrator_environment,
                orchestrator_run_id=self.orchestrator_run_id,
                code_path=self.snapshot.code_path if self.snapshot else None,
                template_id=self.snapshot.template_id
                if self.snapshot
                else None,
                is_templatable=is_templatable,
                trigger_info=trigger_info,
                enable_heartbeat=self.enable_heartbeat,
                exception_info=json.loads(self.exception_info)
                if self.exception_info
                else None,
            )

        resources = None
        if include_resources:
            if self.snapshot:
                source_snapshot = (
                    self.snapshot.source_snapshot.to_model()
                    if self.snapshot.source_snapshot
                    else None
                )
                stack = (
                    self.snapshot.stack.to_model()
                    if self.snapshot.stack
                    else None
                )
                pipeline: Optional["PipelineResponse"] = (
                    self.snapshot.pipeline.to_model()
                )
                build = (
                    self.snapshot.build.to_model()
                    if self.snapshot.build
                    else None
                )
                schedule = (
                    self.snapshot.schedule.to_model()
                    if self.snapshot.schedule
                    else None
                )
                code_reference = (
                    self.snapshot.code_reference.to_model()
                    if self.snapshot.code_reference
                    else None
                )
            else:
                source_snapshot = None
                stack = self.stack.to_model() if self.stack else None
                pipeline = self.pipeline.to_model() if self.pipeline else None
                build = self.build.to_model() if self.build else None
                schedule = self.schedule.to_model() if self.schedule else None
                code_reference = None

            resources = PipelineRunResponseResources(
                user=self.user.to_model() if self.user else None,
                snapshot=self.snapshot.to_model() if self.snapshot else None,
                source_snapshot=source_snapshot,
                stack=stack,
                pipeline=pipeline,
                build=build,
                schedule=schedule,
                code_reference=code_reference,
                model_version=self.model_version.to_model()
                if self.model_version
                else None,
                tags=[tag.to_model() for tag in self.tags],
                log_collection=[log.to_model() for log in self.logs],
                visualizations=[
                    visualization.to_model(
                        include_metadata=False,
                        include_resources=False,
                    )
                    for visualization in self.visualizations
                ],
                trigger=self.trigger.to_model() if self.trigger else None,
                original_run=self.original_run.to_model()
                if self.original_run
                else None,
                active_wait_condition=next(
                    (
                        condition.to_model()
                        for condition in self.wait_conditions
                        if condition.status
                        == RunWaitConditionStatus.PENDING.value
                    ),
                    None,
                ),
            )

        return PipelineRunResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(self, run_update: "PipelineRunUpdate") -> "PipelineRunSchema":
        """Update a `PipelineRunSchema` with a `PipelineRunUpdate`.

        Args:
            run_update: The `PipelineRunUpdate` to update with.

        Raises:
            ValueError: When trying to update the orchestrator run ID of a
                run that already has a different one.

        Returns:
            The updated `PipelineRunSchema`.
        """
        if run_update.orchestrator_run_id:
            if (
                self.orchestrator_run_id
                and self.orchestrator_run_id != run_update.orchestrator_run_id
            ):
                raise ValueError(
                    "Updating the orchestrator run ID of a run with an "
                    "existing orchestrator run ID "
                    f"({self.orchestrator_run_id}) is not allowed."
                )
            self.orchestrator_run_id = run_update.orchestrator_run_id

        if run_update.exception_info:
            self.exception_info = run_update.exception_info.model_dump_json()

        self.updated = utc_now()
        return self

    def update_status(
        self,
        requested_status: Optional[ExecutionStatus],
        status_reason: Optional[str] = None,
    ) -> bool:
        """Update the status of the pipeline run.

        Args:
            requested_status: The requested status of the pipeline run.
            status_reason: The reason for the status of the pipeline run.

        Raises:
            IllegalOperationError: If the requested status transition is
                invalid.

        Returns:
            Whether the status was updated.
        """
        if requested_status in {
            ExecutionStatus.CACHED,
            ExecutionStatus.SKIPPED,
            ExecutionStatus.RETRYING,
            ExecutionStatus.RETRIED,
        }:
            raise IllegalOperationError(
                f"Execution status `{requested_status}` is not valid for "
                "pipeline runs."
            )

        current_status = ExecutionStatus(self.status)

        if (
            requested_status == ExecutionStatus.RESUMING
            and current_status
            not in {
                ExecutionStatus.PAUSED,
                ExecutionStatus.FAILED,
            }
        ):
            raise IllegalOperationError(
                "Only failed or paused runs can be resumed."
            )

        # Snapshot always exists for pipeline runs of newer versions
        assert self.snapshot
        is_dynamic_pipeline = self.snapshot.is_dynamic

        if is_dynamic_pipeline:
            # In dynamic pipelines, the run status is only updated on manual
            # status updates and does not depend on step statuses.
            if requested_status is None and status_reason is None:
                return False

            new_status = requested_status or current_status
        else:
            # For static pipelines we compute the run status based on the step
            # statuses.
            new_status = _compute_static_pipeline_run_status(
                run_status=requested_status or current_status,
                step_statuses=self._get_step_run_statuses(),
                num_steps=self.snapshot.step_count,
            )

        if (
            new_status == ExecutionStatus.PROVISIONING
            and current_status != ExecutionStatus.INITIALIZING
        ):
            # Ignore transitions to provisioning from non-initializing states.
            # This could happen if the orchestrator starts running the pipeline
            # before the client environment can update the status to provisioning.
            return False

        if current_status.is_finished:
            if (
                current_status == ExecutionStatus.FAILED
                and new_status == ExecutionStatus.RESUMING
            ):
                # Allow failed -> resuming transition for retries.
                pass
            elif current_status != new_status:
                raise IllegalOperationError(
                    "Cannot update the status of a finished run."
                )

        self.status = new_status.value

        if is_dynamic_pipeline:
            self.in_progress = not new_status.is_finished
        else:
            self.in_progress = self._check_if_run_in_progress()

        now = utc_now()
        if not self.in_progress and self.end_time is None:
            self.end_time = now
        elif new_status == ExecutionStatus.RESUMING:
            self.end_time = None

        if status_reason:
            self.status_reason = status_reason
        elif (
            current_status == ExecutionStatus.STOPPING
            and new_status == ExecutionStatus.STOPPED
        ):
            # Don't clear status reason for stopping -> stopped transition
            pass
        elif current_status != new_status:
            # Clear status reason when the status changes.
            self.status_reason = None

        self.updated = now
        return True

    def update_placeholder(
        self, request: "PipelineRunRequest"
    ) -> "PipelineRunSchema":
        """Update a placeholder run.

        Args:
            request: The pipeline run request which should replace the
                placeholder.

        Raises:
            RuntimeError: If the DB entry does not represent a placeholder run.
            ValueError: If the run request is not a valid request to replace the
                placeholder run.

        Returns:
            The updated `PipelineRunSchema`.
        """
        if not self.is_placeholder_run():
            raise RuntimeError(
                f"Unable to replace pipeline run {self.id} which is not a "
                "placeholder run."
            )

        if request.is_placeholder_request:
            raise ValueError(
                "Cannot replace a placeholder run with another placeholder run."
            )

        if (
            self.snapshot_id != request.snapshot
            or self.project_id != request.project
        ):
            raise ValueError(
                "Snapshot or project ID of placeholder run "
                "do not match the IDs of the run request."
            )

        if not request.orchestrator_run_id:
            raise ValueError(
                "Orchestrator run ID is required to replace a placeholder run."
            )

        if (
            self.orchestrator_run_id
            and self.orchestrator_run_id != request.orchestrator_run_id
        ):
            raise ValueError(
                "Orchestrator run ID of placeholder run does not match the "
                "ID of the run request."
            )

        orchestrator_environment = json.dumps(request.orchestrator_environment)

        self.orchestrator_run_id = request.orchestrator_run_id
        self.orchestrator_environment = orchestrator_environment
        self.status = request.status.value
        self.in_progress = not request.status.is_finished

        self.updated = utc_now()

        return self

    def is_placeholder_run(self) -> bool:
        """Whether the pipeline run is a placeholder run.

        Returns:
            Whether the pipeline run is a placeholder run.
        """
        return self.status in {
            ExecutionStatus.INITIALIZING.value,
            ExecutionStatus.PROVISIONING.value,
        }

    @staticmethod
    def _step_in_progress(step_id: UUID, session: Session) -> bool:
        from zenml.steps.heartbeat import is_heartbeat_unhealthy
        from zenml.zen_stores.schemas.step_run_schemas import StepRunSchema

        step_run = session.execute(
            select(StepRunSchema).where(StepRunSchema.id == step_id)
        ).scalar_one_or_none()

        if not step_run:
            return False

        status: ExecutionStatus = ExecutionStatus(step_run.status)

        return not is_heartbeat_unhealthy(
            step_run_id=step_run.id,
            status=status,
            heartbeat_threshold=step_run.heartbeat_threshold,
            start_time=step_run.start_time,
            latest_heartbeat=step_run.latest_heartbeat,
        )

    def _check_if_run_in_progress(self) -> bool:
        """Checks whether the run is in progress.

        Raises:
            RuntimeError: If the DB session is missing.

        Returns:
            A flag to indicate whether the run is in progress.
        """
        run_status = ExecutionStatus(self.status)

        if not run_status.is_finished:
            return True

        if run_status == ExecutionStatus.FAILED:
            execution_mode = self.get_pipeline_configuration().execution_mode

            if execution_mode in [
                ExecutionMode.FAIL_FAST,
                ExecutionMode.STOP_ON_FAILURE,
            ]:
                return False

            elif execution_mode == ExecutionMode.CONTINUE_ON_FAILURE:
                from zenml.zen_stores.schemas import StepRunSchema

                if session := object_session(self):
                    step_run_statuses = session.execute(
                        select(
                            StepRunSchema.id,
                            StepRunSchema.name,
                            StepRunSchema.status,
                        )
                        .where(StepRunSchema.pipeline_run_id == self.id)
                        .where(
                            StepRunSchema.status
                            != ExecutionStatus.RETRIED.value
                        )
                    ).all()

                    if self.snapshot and self.snapshot.pipeline_spec:
                        step_dict = self.get_upstream_steps()

                        dag = build_dag(step_dict)

                        step_name_to_id = {
                            name: id_ for id_, name, _ in step_run_statuses
                        }

                        failed_steps = {
                            name
                            for _, name, status in step_run_statuses
                            if ExecutionStatus(status).is_failed
                        }

                        steps_to_skip = set()
                        for failed_step in failed_steps:
                            steps_to_skip.update(
                                find_all_downstream_steps(failed_step, dag)
                            )

                        completed_steps = {
                            name
                            for _, name, status in step_run_statuses
                            if ExecutionStatus(status).is_finished
                        }

                        steps_to_skip.update(
                            failed_steps
                        )  # skip downstream steps
                        steps_to_skip.update(
                            completed_steps
                        )  # skip completed steps

                        steps_statuses = {
                            name: ExecutionStatus(status)
                            for _, name, status in step_run_statuses
                        }

                        for step_name, _ in step_dict.items():
                            if step_name in steps_to_skip:
                                # failed steps downstream
                                continue

                            if step_name not in steps_statuses:
                                # steps that haven't started yet
                                return True

                            if self._step_in_progress(
                                step_name_to_id[step_name], session=session
                            ):
                                # running steps without unhealthy heartbeats
                                return True

                        return False
                    else:
                        in_progress = any(
                            not ExecutionStatus(status).is_finished
                            for _, _, status in step_run_statuses
                        )
                        return in_progress
                else:
                    raise RuntimeError(
                        "Missing DB session to check the in progress "
                        "status of the run."
                    )

        return False


def _compute_static_pipeline_run_status(
    run_status: ExecutionStatus,
    step_statuses: List[ExecutionStatus],
    num_steps: int,
) -> ExecutionStatus:
    """Compute the status of a static pipeline run.

    Args:
        run_status: The status of the run.
        step_statuses: The statuses of the steps in the pipeline run.
        num_steps: The number of steps in the pipeline run.

    Returns:
        The status of the pipeline run.
    """
    if run_status == ExecutionStatus.STOPPING:
        if all(status.is_finished for status in step_statuses):
            return ExecutionStatus.STOPPED
        else:
            return ExecutionStatus.STOPPING
    elif (
        run_status == ExecutionStatus.FAILED
        or ExecutionStatus.FAILED in step_statuses
    ):
        return ExecutionStatus.FAILED
    elif ExecutionStatus.STOPPED in step_statuses:
        if all(status.is_finished for status in step_statuses):
            return ExecutionStatus.STOPPED
        else:
            return ExecutionStatus.STOPPING
    elif (
        ExecutionStatus.RUNNING in step_statuses
        or ExecutionStatus.RETRYING in step_statuses
    ):
        return ExecutionStatus.RUNNING
    elif len(step_statuses) < num_steps:
        if run_status in {
            ExecutionStatus.INITIALIZING,
            ExecutionStatus.PROVISIONING,
        }:
            # The run is still in the initializing or provisioning state but not
            # all steps finished. This is the case for client-side
            # caching/skipping of steps, in which case we do not want to update
            # the run status yet.
            return run_status
        else:
            return ExecutionStatus.RUNNING
    else:
        return ExecutionStatus.COMPLETED
